/*
  Generation Olympia Engine - C++ Port
  Project Home: http://www.sourceforge.net/projects/qb2c/
  
  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.

  $Id: mapedit.cpp,v 1.10 2004/01/20 02:48:43 syn9 Exp $
*/

#include "../engine/fileinput.h"
#include "../engine/engineglobals.h"
#include "../engine/render.h"

#include "mapeditfileinput.h"
#include "mapeditgui.h"
#include "mapeditglobals.h"
#include "mapedit.h"

#include "../common/strutil.h"

#include <math.h>

void applyrot()
{
	unsigned int i;
	float nsx, nsy, nsz;

	for (i = 0; i < pointinfo.size(); i++)
	{
		nsx = pointinfo[i].nsx;
		nsy = pointinfo[i].nsy;
		nsz = pointinfo[i].nsz;

		pointinfo[i].nsx = nsx * prexx + nsy * preyx + nsz * prezx;
		pointinfo[i].nsy = nsx * prexy + nsy * preyy + nsz * prezy;
		pointinfo[i].nsz = nsx * prexz + nsy * preyz + nsz * prezz;
	}
}


void calcsprites()
{
    short ind, sox, scx, scy;
    float nx, ny, nz, sx, sy, sz, ratio;

    nx = 1 + playerx - focusofs[0];
    ny = playery - focusofs[1];
    nz = playerz + .5 - focusofs[2];

    sx = nx * prexx + ny * preyx + nz * prezx - camx;
    sz = nx * prexz + ny * preyz + nz * prezz - camz;

    sox = short(xcenter + sx / sz * xzoomratio);
   
    nx = playerx - focusofs[0];
    ny = playery - focusofs[1];
    nz = playerz + .5 - focusofs[2];

    sx = nx * prexx + ny * preyx + nz * prezx - camx;
    sy = nx * prexy + ny * preyy + nz * prezy - camy;
    sz = nx * prexz + ny * preyz + nz * prezz - camz;

    scx = short(xcenter + sx / sz * xzoomratio);
    scy = short(ycenter + sy / sz * yzoomratio - scrny);

    ratio = (float)abs(scx - sox) / 22;

    scx = scx - int(16 * ratio);
    scy = scy - int(30 * ratio);

    ind = -int(sz * 75) - 400;

    for(;;)
    {
        if (ind > vdist || ind < 0)
            break;

        if (dbuffer[ind] > 0)
            ind = ind + 1;
        if (dbuffer[ind] == 0)
        {
            dbuffer[ind] = 1;
            if (ind < start)
                start = ind;
            if (ind > depth)
                depth = ind;
            break;
        }
    }

    pscx = scx;
    pscy = scy;
    pratio = ratio;

    
}

void checkkbd()
{

	short a, b, c, p;
    
	if (key[KEY_G])
	{
		generating_radiosity = TRUE;
		bk_camx = camx;
		bk_camy = camy;
		bk_camz = camz;
		bk_camrx = camrx;
		bk_camry = camry;
		bk_camrz = camrz;
		bk_scrny = scrny;
		bk_gridbase = gridbase;
		bk_xzoomratio = xzoomratio;
		bk_yzoomratio = yzoomratio;
		bk_bfc = bfc;

		camx = 0;
		camy = 0;
		camz = 0;
		camrx = 0;
		camry = 0;
		camrz = 0;
		scrny = 0;
		gridbase = 0;
		stp = 4;
		
		bfc = FALSE;

		
		editingmode = 0;

		for (a = 0; a <= endplane; a++)
		{
			if (planeinfo[a].active)
			{
				for (b = 0; b <= planeinfo[a].lastpoly; b++)
				{
					if (planeinfo[a].polyinfo[b][6] == 0 && planeinfo[a].polyinfo[b][8] == 0)
					{
					
						rectfill(planeinfo[a].radiosity[b], 0, 0, 15, 15, 0);												
					}
				}
			}
		}

		return;
	}

	
	
	if (key[KEY_F5] && freeze < timer) 
	{		
		hlplane = !hlplane;
		freeze = timer + .002 * fpsr;
	}
	if (key[KEY_F6] && freeze < timer) 
	{		
		hlpoly = !hlpoly;
		nselpoly = 0;
		freeze = timer + .002 * fpsr;
	}
	if (key[KEY_F7] && freeze < timer) 
	{		
		hlpoint = !hlpoint;
		nselpoint = 0;
		freeze = timer + .002 * fpsr;
		editingmode = select_point;
	}

	if (hlplane)
	{
		hlpoint = FALSE;
		nselpoint = 0;
		hlpoly = FALSE;
		nselpoly = 0;
	}
	if (hlpoly)
	{
		hlpoint = FALSE;
		nselpoint = 0;
		hlplane = FALSE;
		if (key[KEY_A])
		{
			nselpoly = 0;
			for (c = 0; c <= planeinfo[selplane].lastpoly; c++)
			{
				if (planeinfo[selplane].polyinfo[c][6] < 2)
				{
					selpoly[nselpoly] = c;
					nselpoly++;
				}
			}
		}
	}
	if (hlpoint)
	{
		hlplane = FALSE;		
		hlpoly = FALSE;
		nselpoly = 0;
	}
	

	if (key[KEY_F9] && freeze < timer)
	{
		grid = !grid;		
		freeze = timer + .005 * fpsr;
	}

	if (key[KEY_F10] && freeze < timer)
	{
		fillmode = 1 - fillmode;
		bfc = !bfc;
		freeze = timer + .005 * fpsr;
	}

	if (key[KEY_F11]) save_bmp("shot.bmp", screen, pal);

	if (key[KEY_E])	camz = camz - 2;		
	if (key[KEY_D])	camz = camz + 2;
	if (key[KEY_W])	camy = camy - 1;
	if (key[KEY_R])	camy = camy + 1;
	if (key[KEY_S])	
	{
		camrx = camrx - 1;
		calcsteps();
		setup_scene();
	}
	if (key[KEY_F])	
	{
		camrx = camrx + 1;
		calcsteps();
		setup_scene();
	}


	if (key[KEY_LEFT]) fightfocus[0] = fightfocus[0] - .5;
	if (key[KEY_RIGHT]) fightfocus[0] = fightfocus[0] + .5;
	if (key[KEY_UP]) fightfocus[2] = fightfocus[2] - .5;
	if (key[KEY_DOWN]) fightfocus[2] = fightfocus[2] + .5;

	if (key[KEY_MINUS_PAD] && freeze < timer)
	{
		gridbase-= gridy;
		fightfocus[1] = gridbase;
		freeze = timer + .001 * fpsr;
	}
	if (key[KEY_PLUS_PAD] && freeze < timer)
	{
		gridbase+= gridy;
		fightfocus[1] = gridbase;
		freeze = timer + .001 * fpsr;
	}

	float spd;

	if (editingmode == modify_plane_rotate)
	{		
		if (key[KEY_2_PAD] && freeze < timer) 
		{
			spd = 1;
			if (key[KEY_SPACE]) spd = 15;
			planeinfo[selplane].xrot = u(planeinfo[selplane].xrot + spd);
			rotate_plane(selplane);
			freeze = timer + .001 * fpsr;
		}
		if (key[KEY_8_PAD] && freeze < timer)
		{
			spd = 1;
			if (key[KEY_SPACE]) spd = 15;
			planeinfo[selplane].xrot = u(planeinfo[selplane].xrot - spd);
			rotate_plane(selplane);
			freeze = timer + .001 * fpsr;
		}
		if (key[KEY_4_PAD] && freeze < timer) 
		{
			spd = 1;
			if (key[KEY_SPACE]) spd = 15;
			planeinfo[selplane].yrot = u(planeinfo[selplane].yrot + spd);
			rotate_plane(selplane);
			freeze = timer + .001 * fpsr;
		}
		if (key[KEY_6_PAD] && freeze < timer)
		{
			spd = 1;
			if (key[KEY_SPACE]) spd = 15;
			planeinfo[selplane].yrot = u(planeinfo[selplane].yrot - spd);
			rotate_plane(selplane);
			freeze = timer + .001 * fpsr;
		}
		if (key[KEY_1_PAD] && freeze < timer) 
		{
			spd = 1;
			if (key[KEY_SPACE]) spd = 15;
			planeinfo[selplane].zrot = u(planeinfo[selplane].zrot + spd);
			rotate_plane(selplane);
			freeze = timer + .001 * fpsr;
		}
		if (key[KEY_7_PAD] && freeze < timer)
		{
			spd = 1;
			if (key[KEY_SPACE]) spd = 15;
			planeinfo[selplane].zrot = u(planeinfo[selplane].zrot - spd);
			rotate_plane(selplane);
			freeze = timer + .001 * fpsr;
		}
		if (key[KEY_0_PAD] && freeze < timer)
		{			
			planeinfo[selplane].xrot = 0;
			planeinfo[selplane].yrot = 0;
			planeinfo[selplane].zrot = 0;
			rotate_plane(selplane);
		}
	}

	if (editingmode == modify_point_rotate)
	{		
		if (key[KEY_2_PAD] && freeze < timer) 
		{
			spd = 1;
			if (key[KEY_SPACE]) spd = 15;
			planeinfo[selplane].xrot = u(planeinfo[selplane].xrot + spd);
			rotate_point(selplane);
			freeze = timer + .001 * fpsr;
		}
		if (key[KEY_8_PAD] && freeze < timer)
		{
			spd = 1;
			if (key[KEY_SPACE]) spd = 15;
			planeinfo[selplane].xrot = u(planeinfo[selplane].xrot - spd);
			rotate_point(selplane);
			freeze = timer + .001 * fpsr;
		}
		if (key[KEY_4_PAD] && freeze < timer) 
		{
			spd = 1;
			if (key[KEY_SPACE]) spd = 15;
			planeinfo[selplane].yrot = u(planeinfo[selplane].yrot + spd);
			rotate_point(selplane);
			freeze = timer + .001 * fpsr;
		}
		if (key[KEY_6_PAD] && freeze < timer)
		{
			spd = 1;
			if (key[KEY_SPACE]) spd = 15;
			planeinfo[selplane].yrot = u(planeinfo[selplane].yrot - spd);
			rotate_point(selplane);
			freeze = timer + .001 * fpsr;
		}
		if (key[KEY_1_PAD] && freeze < timer) 
		{
			spd = 1;
			if (key[KEY_SPACE]) spd = 15;
			planeinfo[selplane].zrot = u(planeinfo[selplane].zrot + spd);
			rotate_point(selplane);
			freeze = timer + .001 * fpsr;
		}
		if (key[KEY_7_PAD] && freeze < timer)
		{
			spd = 1;
			if (key[KEY_SPACE]) spd = 15;
			planeinfo[selplane].zrot = u(planeinfo[selplane].zrot - spd);
			rotate_point(selplane);
			freeze = timer + .001 * fpsr;
		}
		if (key[KEY_0_PAD] && freeze < timer)
		{			
			planeinfo[selplane].xrot = 0;
			planeinfo[selplane].yrot = 0;
			planeinfo[selplane].zrot = 0;
			rotate_point(selplane);
		}
	}

	float trans[3];
	if (editingmode == modify_plane_translate_free && freeze < timer)
	{
		trans[0] = 0;
		trans[1] = 0;
		trans[2] = 0;
		spd = .1;
		if (key[KEY_SPACE]) spd = .5;

		if (key[KEY_2_PAD])
		{
			trans[0] = 0;
			trans[1] = 0;
			trans[2] = spd;
			freeze = timer + .001 * fpsr;
		}
		if (key[KEY_8_PAD])
		{
			trans[0] = 0;
			trans[1] = 0;
			trans[2] = -spd;
			freeze = timer + .001 * fpsr;
		}
		if (key[KEY_4_PAD])
		{
			trans[0] = -spd;
			trans[1] = 0;
			trans[2] = 0;
			freeze = timer + .001 * fpsr;
		}
		if (key[KEY_6_PAD])
		{
			trans[0] = spd;
			trans[1] = 0;
			trans[2] = 0;
			freeze = timer + .001 * fpsr;
		}
		if (key[KEY_1_PAD])
		{
			trans[0] = 0;
			trans[1] = -spd;
			trans[2] = 0;
			freeze = timer + .001 * fpsr;
		}
		if (key[KEY_7_PAD])
		{
			trans[0] = 0;
			trans[1] = spd;
			trans[2] = 0;
			freeze = timer + .001 * fpsr;
		}

		planeinfo[selplane].basex += trans[0];
		planeinfo[selplane].basey += trans[1];
		planeinfo[selplane].basez += trans[2];
		
		rotate_plane(selplane);
	}


	
	if (editingmode == modify_point_translate_free && freeze < timer)
	{
		trans[0] = 0;
		trans[1] = 0;
		trans[2] = 0;
		spd = .1;
		if (key[KEY_SPACE]) spd = .5;

		if (key[KEY_2_PAD])
		{
			trans[0] = 0;
			trans[1] = 0;
			trans[2] = spd;
			freeze = timer + .001 * fpsr;
		}
		if (key[KEY_8_PAD])
		{
			trans[0] = 0;
			trans[1] = 0;
			trans[2] = -spd;
			freeze = timer + .001 * fpsr;
		}
		if (key[KEY_4_PAD])
		{
			trans[0] = -spd;
			trans[1] = 0;
			trans[2] = 0;
			freeze = timer + .001 * fpsr;
		}
		if (key[KEY_6_PAD])
		{
			trans[0] = spd;
			trans[1] = 0;
			trans[2] = 0;
			freeze = timer + .001 * fpsr;
		}
		if (key[KEY_1_PAD])
		{
			trans[0] = 0;
			trans[1] = -spd;
			trans[2] = 0;
			freeze = timer + .001 * fpsr;
		}
		if (key[KEY_7_PAD])
		{
			trans[0] = 0;
			trans[1] = spd;
			trans[2] = 0;
			freeze = timer + .001 * fpsr;
		}

		for (c = 0; c < nselpoint; c++)
		{
			p = (short)selpoint[c][3];
			if (p > -1)
			{				
				planeinfo[selplane].pointinfo[p][0] += trans[0];
				planeinfo[selplane].pointinfo[p][1] += trans[1];
				planeinfo[selplane].pointinfo[p][2] += trans[2];
				selpoint[c][0] += trans[0];
				selpoint[c][1] += trans[1];
				selpoint[c][2] += trans[2];
			}
		}
		
		rotate_plane(selplane);
	}

	
}

int	checkmouse(short x1, short y1, short x2, short y2)
{
	if (mouse_x >= x1 && mouse_y >= y1 && mouse_x <= x2 && mouse_y <= y2)
	{
		return -1;
	} else {
		return 0;
	}
}

bool checktolerance(float x1, float y1, float z1, float x2, float y2, float z2)
{
	short pass = 0;

	if (x1 > x2 - tolerance && x1 < x2 + tolerance) pass++;
	if (y1 > y2 - tolerance && y1 < y2 + tolerance) pass++;
	if (z1 > z2 - tolerance && z1 < z2 + tolerance) pass++;


	if (pass == 3) return TRUE;
	return FALSE;
}

void clear_selected()
{
	hlplane = FALSE;
	hlpoly = FALSE;
	hlpoint = FALSE;
	nselpoly = 0;
	nselpoint = 0;
}

void condense_rotate(short plane)
{
	int c;
	for (c = 0; c <= planeinfo[plane].lastpoint; c++)
	{
		planeinfo[plane].pointinfo[c][0] = planeinfo[plane].pointinfo[c][3];
		planeinfo[plane].pointinfo[c][1] = planeinfo[plane].pointinfo[c][4];
		planeinfo[plane].pointinfo[c][2] = planeinfo[plane].pointinfo[c][5];
	}
	planeinfo[plane].xrot = 0;
	planeinfo[plane].yrot = 0;
	planeinfo[plane].zrot = 0;
}

void drawhud()
{
	
	char fpstext[100];
	char vistext[100];
	char tmptxt[100];

	int whitec = makecol(192,192,192), bluec = makecol(0,0,192), greenc = makecol(0,192,0), orangec = makecol(224, 224, 32);
	int ofs, c, b, i, a, p;
	int newplane;

	float nx, ny, nz, sx, sy, sz, x, y, z;

	
	sprintf(fpstext, "%d fps", FPS);
	sprintf(vistext, "%d vis polys", vispolys);

	win(0, 0, 0, 127, 23);
	
	text_mode(-1);
	textout(offscreen, font, fpstext, 4, 4, whitec);
	textout(offscreen, font, vistext, 4, 12, whitec);

	if (hud)
	{
	win(0, xres - 192, 0, xres - 1, 5+6*8);
	sprintf(tmptxt, "Scene Inforamtion");
	textout(offscreen, font, tmptxt, xres-192 + 4, 2, whitec);
	sprintf(tmptxt, "nPoints %d", pointinfo.size());
	textout(offscreen, font, tmptxt, xres-192 + 4, 10, whitec);
	sprintf(tmptxt, "nPolys  %d", polyinfo.size());
	textout(offscreen, font, tmptxt, xres-192 + 4, 18, whitec);
	sprintf(tmptxt, "nPlanes %d", nplanes);
	textout(offscreen, font, tmptxt, xres-192 + 4, 26, whitec);
	sprintf(tmptxt, "nLights %d", nlights);	
	textout(offscreen, font, tmptxt, xres-192 + 4, 2 + 4 * 8, whitec);

	win(0, xres - 192, 6+6*8, xres - 1, 6+6*8+5+15*8);
	sprintf(tmptxt, "Plane Information");
	textout(offscreen, font, tmptxt, xres-192 + 4, 6+6*8+2 + 0*8, whitec);
	sprintf(tmptxt, "Base X %f", planeinfo[selplane].basex);
	textout(offscreen, font, tmptxt, xres-192 + 4, 6+6*8+2 + 1*8, whitec);
	sprintf(tmptxt, "Base Y %f", planeinfo[selplane].basey);
	textout(offscreen, font, tmptxt, xres-192 + 4, 6+6*8+2 + 2*8, whitec);
	sprintf(tmptxt, "Base Z %f", planeinfo[selplane].basez);
	textout(offscreen, font, tmptxt, xres-192 + 4, 6+6*8+2 + 3*8, whitec);
	sprintf(tmptxt, "End X %f", planeinfo[selplane].endx);
	textout(offscreen, font, tmptxt, xres-192 + 4, 6+6*8+2 + 4*8, whitec);
	sprintf(tmptxt, "End Y %f", planeinfo[selplane].endy);
	textout(offscreen, font, tmptxt, xres-192 + 4, 6+6*8+2 + 5*8, whitec);
	sprintf(tmptxt, "End Z %f", planeinfo[selplane].endz);
	textout(offscreen, font, tmptxt, xres-192 + 4, 6+6*8+2 + 6*8, whitec);
	sprintf(tmptxt, "X Rot %f", planeinfo[selplane].xrot);
	textout(offscreen, font, tmptxt, xres-192 + 4, 6+6*8+2 + 7*8, whitec);
	sprintf(tmptxt, "Y Rot %f", planeinfo[selplane].yrot);
	textout(offscreen, font, tmptxt, xres-192 + 4, 6+6*8+2 + 8*8, whitec);
	sprintf(tmptxt, "Z Rot %f", planeinfo[selplane].zrot);
	textout(offscreen, font, tmptxt, xres-192 + 4, 6+6*8+2 + 9*8, whitec);
	sprintf(tmptxt, "Height %f", planeinfo[selplane].height);
	textout(offscreen, font, tmptxt, xres-192 + 4, 6+6*8+2 + 10*8, whitec);
	sprintf(tmptxt, "Cols %d", planeinfo[selplane].cols);
	textout(offscreen, font, tmptxt, xres-192 + 4, 6+6*8+2 + 11*8, whitec);
	sprintf(tmptxt, "Rows %d", planeinfo[selplane].rows);
	textout(offscreen, font, tmptxt, xres-192 + 4, 6+6*8+2 + 12*8, whitec);
	sprintf(tmptxt, "Points %d", planeinfo[selplane].npoints);
	textout(offscreen, font, tmptxt, xres-192 + 4, 6+6*8+2 + 13*8, whitec);
	sprintf(tmptxt, "Polys  %d", planeinfo[selplane].npolies);
	textout(offscreen, font, tmptxt, xres-192 + 4, 6+6*8+2 + 14*8, whitec);
	
	

	win(0, xres - 192 -140, 0, xres - 192 - 1, 5+7 * 8);
	sprintf(tmptxt, "Camera");
	textout(offscreen, font, tmptxt, xres-192 + 4 -140, 2, whitec);
	sprintf(tmptxt, "camx %f", camx);
	textout(offscreen, font, tmptxt, xres-192 + 4 -140, 10, whitec);
	sprintf(tmptxt, "camy %f", camy);
	textout(offscreen, font, tmptxt, xres-192 + 4 -140, 18, whitec);
	sprintf(tmptxt, "camz %f", camz);
	textout(offscreen, font, tmptxt, xres-192 + 4 -140, 26, whitec);	
	sprintf(tmptxt, "camrx %f", camrx);
	textout(offscreen, font, tmptxt, xres-192 + 4 -140, 34, whitec);
	sprintf(tmptxt, "camry %f", camry);
	textout(offscreen, font, tmptxt, xres-192 + 4 -140, 42, whitec);
	sprintf(tmptxt, "camrz %f", camrz);
	textout(offscreen, font, tmptxt, xres-192 + 4 -140, 50, whitec);
	
	win(0, xres - 192 -140, 6+7*8, xres - 192 - 1, 6+7*8+5+4*8);
	sprintf(tmptxt, "Focus Position");
	textout(offscreen, font, tmptxt, xres-192 + 4 -140, 6+7*8+2, whitec);
	sprintf(tmptxt, "x %f", fightfocus[0]);
	textout(offscreen, font, tmptxt, xres-192 + 4 -140, 6+7*8+10, whitec);
	sprintf(tmptxt, "y %f", fightfocus[1]);
	textout(offscreen, font, tmptxt, xres-192 + 4 -140, 6+7*8+18, whitec);
	sprintf(tmptxt, "z %f", fightfocus[2]);
	textout(offscreen, font, tmptxt, xres-192 + 4 -140, 6+7*8+26, whitec);
	
	win(0, xres - 192 -140 -140, 0, xres - 192 - 1 - 140, 5+12 * 8);
	sprintf(tmptxt, "Selection");
	textout(offscreen, font, tmptxt, xres-192 + 4 -140 -140, 2+0*8, whitec);
	sprintf(tmptxt, "Plane %d", selplane);
	textout(offscreen, font, tmptxt, xres-192 + 4 -140 -140, 2+1*8, whitec);
	sprintf(tmptxt, "Poly  %d", nselpoly);
	textout(offscreen, font, tmptxt, xres-192 + 4 -140 -140, 2+2*8, whitec);		
	sprintf(tmptxt, "Point %d", nselpoint);
	textout(offscreen, font, tmptxt, xres-192 + 4 -140 -140, 2+3*8, whitec);
	sprintf(tmptxt, "Hovering Over");
	textout(offscreen, font, tmptxt, xres-192 + 4 -140 -140, 2+5*8, whitec);
	sprintf(tmptxt, "Plane %d", tselplane);
	textout(offscreen, font, tmptxt, xres-192 + 4 -140 -140, 2+6*8, whitec);
	sprintf(tmptxt, "Poly  %d", tselpoly);
	textout(offscreen, font, tmptxt, xres-192 + 4 -140 -140, 2+7*8, whitec);		
	sprintf(tmptxt, "Point %d", tselpoint);
	textout(offscreen, font, tmptxt, xres-192 + 4 -140 -140, 2+8*8, whitec);
	sprintf(tmptxt, " X %f", tselpointinfo[0]);
	textout(offscreen, font, tmptxt, xres-192 + 4 -140 -140, 2+9*8, whitec);
	sprintf(tmptxt, " Y %f", tselpointinfo[1]);
	textout(offscreen, font, tmptxt, xres-192 + 4 -140 -140, 2+10*8, whitec);
	sprintf(tmptxt, " Z %f", tselpointinfo[2]);
	textout(offscreen, font, tmptxt, xres-192 + 4 -140 -140, 2+11*8, whitec);	
	
	if (hlpoint)
	{
		ofs = 6+7*8+5+4*8;		
		win(0, xres - 192 -140, ofs, xres - 192 - 1, ofs+5+(1+5*nselpoint)*8);
		sprintf(tmptxt, "Sel Points");
		textout(offscreen, font, tmptxt, xres-192 + 4 -140, ofs+2, whitec);
		for (c = 0; c < nselpoint; c++)
		{			
			sprintf(tmptxt, "%d:X %f", c,selpoint[c][0]);
			textout(offscreen, font, tmptxt, xres-192 + 4 -140, ofs+2+(1+5*c+0)*8, whitec);
			sprintf(tmptxt, "  Y %f", selpoint[c][1]);
			textout(offscreen, font, tmptxt, xres-192 + 4 -140, ofs+2+(1+5*c+1)*8, whitec);
			sprintf(tmptxt, "  Z %f", selpoint[c][2]);
			textout(offscreen, font, tmptxt, xres-192 + 4 -140, ofs+2+(1+5*c+2)*8, whitec);
			sprintf(tmptxt, "  P %f", selpoint[c][3]);
			textout(offscreen, font, tmptxt, xres-192 + 4 -140, ofs+2+(1+5*c+3)*8, whitec);
		}
	}

	}

	if (editingmode >= sys && editingmode < 100)
	{
		win(1, 0, 24 + 12 * 0, 127, 35 + 12 * 0);
	} else {
		win(0, 0, 24 + 12 * 0, 127, 35 + 12 * 0);
	}
	sprintf(tmptxt, "ESC System");
	textout(offscreen, font, tmptxt, 4, 24 + 12 * 0 + 2, whitec);
	if (editingmode != sys)
	{
		if ((mouse_b & 1 && checkmouse(0, 24 + 12 * 0, 127, 35 + 12 * 0)) || key[KEY_ESC])
		{
			editingmode = sys;
			clear_selected();
		}
	}

	if (editingmode >= sys && editingmode < 100)
	{
		if (editingmode == sys || editingmode >= system_save && editingmode < system_load)
		{
			if (editingmode == system_save) win(1, 128, 24 + 12 * 0, 255, 35 + 12 * 0);
			if (editingmode != system_save) win(0, 128, 24 + 12 * 0, 255, 35 + 12 * 0);
			sprintf(tmptxt, "1 Save");
			textout(offscreen, font, tmptxt, 132, 24 + 12 * 0 + 2, whitec);
		}
		
		if (editingmode == sys || editingmode >= system_load && editingmode < 030)
		{
			if (editingmode == system_load) win(1, 128, 24 + 12 * 1, 255, 35 + 12 * 1);
			if (editingmode != system_load) win(0, 128, 24 + 12 * 1, 255, 35 + 12 * 1);
			sprintf(tmptxt, "2 Load");
			textout(offscreen, font, tmptxt, 132, 24 + 12 * 1 + 2, whitec);
		}
				
		if (editingmode == sys || editingmode >= system_import && editingmode < 040)
		{
			if (editingmode == system_import) win(1, 128, 24 + 12 * 2, 255, 35 + 12 * 2);
			if (editingmode != system_import) win(0, 128, 24 + 12 * 2, 255, 35 + 12 * 2);
			sprintf(tmptxt, "3 Import");
			textout(offscreen, font, tmptxt, 132, 24 + 12 * 2 + 2, whitec);
		}
		
		if (editingmode == sys || editingmode >= system_preferences && editingmode < 050)
		{
			if (editingmode == system_preferences) win(1, 128, 24 + 12 * 3, 255, 35 + 12 * 3);
			if (editingmode != system_preferences) win(0, 128, 24 + 12 * 3, 255, 35 + 12 * 3);
			sprintf(tmptxt, "4 Preferences");
			textout(offscreen, font, tmptxt, 132, 24 + 12 * 3 + 2, whitec);
		}
		
		if (editingmode == sys && freeze < timer)
		{
			if ((mouse_b & 1 && checkmouse(128, 24 + 12 * 0, 255, 35 + 12 * 0)) || key[KEY_1])
			{
				editingmode = system_save;
				freeze = timer + .002 * fpsr;
			}
			if ((mouse_b & 1 && checkmouse(128, 24 + 12 * 1, 255, 35 + 12 * 1)) || key[KEY_2]) editingmode = system_load;
			if ((mouse_b & 1 && checkmouse(128, 24 + 12 * 2, 255, 35 + 12 * 2)) || key[KEY_3])
			{
				editingmode = system_import;
				freeze = timer + .002 * fpsr;
			}
			if ((mouse_b & 1 && checkmouse(128, 24 + 12 * 3, 255, 35 + 12 * 3)) || key[KEY_4]) editingmode = system_preferences;
		}

		if (editingmode == system_save)
		{			
			win(0, 256, 24 + 12 * 0, 383, 35 + 12 * 0);
			sprintf(tmptxt, "1 Detailed");
			textout(offscreen, font, tmptxt, 260, 24 + 12 * 0 + 2, whitec);
						
			win(0, 256, 24 + 12 * 1, 383, 35 + 12 * 1);
			sprintf(tmptxt, "2 Simple");
			textout(offscreen, font, tmptxt, 260, 24 + 12 * 1 + 2, whitec);			
			
			if (freeze < timer)
			{
				if ((mouse_b & 1 && checkmouse(256, 24 + 12 * 0, 383, 35 + 12 * 0)) || key[KEY_1]) editingmode = system_save_detailed;
				if ((mouse_b & 1 && checkmouse(256, 24 + 12 * 1, 383, 35 + 12 * 1)) || key[KEY_2]) editingmode = system_save_simple;
			}
		}

		if (editingmode == system_import)
		{
			win(0, 256, 24 + 12 * 2, 383, 35 + 12 * 2);
			sprintf(tmptxt, "1 Simple");
			textout(offscreen, font, tmptxt, 260, 24 + 12 * 2 + 2, whitec);
						
			win(0, 256, 24 + 12 * 3, 383, 35 + 12 * 3);
			sprintf(tmptxt, "2 ASE");
			textout(offscreen, font, tmptxt, 260, 24 + 12 * 3 + 2, whitec);			
			
			if (freeze < timer)
			{
				if ((mouse_b & 1 && checkmouse(256, 24 + 12 * 2, 383, 35 + 12 * 2)) || key[KEY_1]) editingmode = system_save_detailed;
				if ((mouse_b & 1 && checkmouse(256, 24 + 12 * 3, 383, 35 + 12 * 3)) || key[KEY_2]) editingmode = system_save_simple;
			}
		}
	}

	if (editingmode == addplane)
	{
		win(1, 0, 24 + 12 * 1, 127, 35 + 12 * 1);
	} else {
		win(0, 0, 24 + 12 * 1, 127, 35 + 12 * 1);
	}
	sprintf(tmptxt, "F1 Add Plane");
	textout(offscreen, font, tmptxt, 4, 24 + 12 * 1 + 2, whitec);
	if (editingmode != addplane)
	{
		if ((mouse_b & 1 && checkmouse(0, 24 + 12 * 1, 127, 35 + 12 * 1)) || key[KEY_F1])
		{
			clear_selected();
			editingmode = addplane;
			hlpoint = TRUE;
			nselpoint = 0;
		}

	}

	if (editingmode >= addpoly && editingmode < 300)
	{
		win(1, 0, 24 + 12 * 2, 127, 35 + 12 * 2);
	} else {
		win(0, 0, 24 + 12 * 2, 127, 35 + 12 * 2);
	}
	sprintf(tmptxt, "F2 Add Poly");
	textout(offscreen, font, tmptxt, 4, 24 + 12 * 2 + 2, whitec);
	if (editingmode != addpoly)
	{
		if ((mouse_b & 1 && checkmouse(0, 24 + 12 * 2, 127, 35 + 12 * 2)) || key[KEY_F2])
		{
			editingmode = addpoly;
			freeze = timer + .002 * fpsr;
			clear_selected();
			hlpoint = TRUE;
			nselpoint = 0;
		}
	}

	if (editingmode >= addlight && editingmode < 400)
	{
		win(1, 0, 24 + 12 * 3, 127, 35 + 12 * 3);
	} else {
		win(0, 0, 24 + 12 * 3, 127, 35 + 12 * 3);
	}
	sprintf(tmptxt, "F3 Add Light");
	textout(offscreen, font, tmptxt, 4, 24 + 12 * 3 + 2, whitec);
	if (editingmode != addlight)
	{
		if ((mouse_b & 1 && checkmouse(0, 24 + 12 * 3, 127, 35 + 12 * 3)) || key[KEY_F3])
		{
			editingmode = addlight;
			clear_selected();
			hlpoint = TRUE;				
			freeze = timer + .002 * fpsr;
		}
	}

	if (pointinfo.size() > 0)
	{
		if (editingmode >= 400 && editingmode < 500)
		{	
			win(1, 0, 24 + 12 * 4, 127, 35 + 12 * 4);
		} else {
			win(0, 0, 24 + 12 * 4, 127, 35 + 12 * 4);
		}
		sprintf(tmptxt, "F4 Remove");
		textout(offscreen, font, tmptxt, 4, 24 + 12 * 4 + 2, whitec);
		if (editingmode != remov && pointinfo.size() > 0)
		{
			if ((mouse_b & 1 && checkmouse(0, 24 + 12 * 4, 127, 35 + 12 * 4)) || key[KEY_F4])
			{
				editingmode = remov;
				freeze = timer + .002 * fpsr;
			}
		}
	}

	if (editingmode >= 400 && editingmode < 500)
	{
		if (editingmode == remov || editingmode >= remove_plane && editingmode < remove_poly)
		{
			if (editingmode == remove_plane) win(1, 128, 24 + 12 * 4, 255, 35 + 12 * 4);
			if (editingmode != remove_plane) win(0, 128, 24 + 12 * 4, 255, 35 + 12 * 4);
			sprintf(tmptxt, "1 Plane");
			textout(offscreen, font, tmptxt, 132, 24 + 12 * 4 + 2, whitec);
		}
		
		if (editingmode == remov && nselpoly > 0 || editingmode >= remove_poly && editingmode < 430 && nselpoly > 0)
		{
			if (editingmode == remove_poly) win(1, 128, 24 + 12 * 5, 255, 35 + 12 * 5);
			if (editingmode != remove_poly) win(0, 128, 24 + 12 * 5, 255, 35 + 12 * 5);
			sprintf(tmptxt, "2 Poly");
			textout(offscreen, font, tmptxt, 132, 24 + 12 * 5 + 2, whitec);
		}
				
		if (editingmode == remov || editingmode >= remove_isolated_points && editingmode < 440)
		{
			if (editingmode == remove_isolated_points) win(1, 128, 24 + 12 * 6, 269, 35 + 12 * 6);
			if (editingmode != remove_isolated_points) win(0, 128, 24 + 12 * 6, 269, 35 + 12 * 6);
			sprintf(tmptxt, "3 Isolated Points");
			textout(offscreen, font, tmptxt, 132, 24 + 12 * 6 + 2, whitec);
		}

		if (editingmode == remov && nselpoint > 0 || editingmode >= remove_light && editingmode < 450 && nselpoint > 0)
		{
			if (editingmode == remove_light) win(1, 128, 24 + 12 * 7, 269, 35 + 12 * 7);
			if (editingmode != remove_light) win(0, 128, 24 + 12 * 7, 269, 35 + 12 * 7);
			sprintf(tmptxt, "4 Light");
			textout(offscreen, font, tmptxt, 132, 24 + 12 * 7 + 2, whitec);
		}
		
		if (editingmode == remov && freeze < timer)
		{
			if ((mouse_b & 1 && checkmouse(128, 24 + 12 * 4, 255, 35 + 12 * 4)) || key[KEY_1])
			{
				editingmode = 0;
				planeinfo[selplane].active = FALSE;
				nplanes--;

				selplane = 0;
				for (c = 0; c < max_planes; c++)
				{
					if (planeinfo[c].active)
					{							
						selplane = c;
						continue;
					}
				}

				endplane = 0;
				for (c = 0; c < max_planes; c++)
				{
					if (planeinfo[c].active)
					{							
						newplane = c;
						if (newplane > endplane) endplane = newplane;
					}
				}

				freeze = timer + .002 * fpsr;
				setup_scene();
			}
			if ((mouse_b & 1 && checkmouse(128, 24 + 12 * 5, 255, 35 + 12 * 5) && nselpoly > 0) || key[KEY_2] && nselpoly > 0)
			{
				for (c = 0; c < nselpoly; c++)
				{
					planeinfo[selplane].polyinfo[selpoly[c]][6] = 2;
				}
				nselpoly = 0;				
				editingmode = 0;
				freeze = timer + .002 * fpsr;
				planeinfo[selplane].condensed = TRUE;

				condense_rotate(selplane);
				rotate_plane(selplane);
			}
			if ((mouse_b & 1 && checkmouse(128, 24 + 12 * 6, 269, 35 + 12 * 6)) || key[KEY_3]) editingmode = remove_isolated_points;
			if ((mouse_b & 1 && checkmouse(128, 24 + 12 * 7, 269, 35 + 12 * 7) && nselpoint > 0) || key[KEY_4] && nselpoint > 0)
			{
				editingmode = remove_light;
			}
		}
	}

	if(pointinfo.size() > 0)
	{
		if (editingmode >= modify && editingmode < 900)
		{	
			win(1, 0, 24 + 12 * 8, 127, 35 + 12 * 8);
		} else {
			win(0, 0, 24 + 12 * 8, 127, 35 + 12 * 8);
		}
		sprintf(tmptxt, "F8 Modify");
		textout(offscreen, font, tmptxt, 4, 24 + 12 * 8 + 2, whitec);
		if (editingmode != modify )
		{
			if ((mouse_b & 1 && checkmouse(0, 24 + 12 * 8, 127, 35 + 12 * 8)) || key[KEY_F8])
			{
				editingmode = modify;
				freeze = timer + .002 * fpsr;				
			}
		}
	}

	if (editingmode >= 800 && editingmode < 900)
	{
		if (editingmode == modify || editingmode >= modify_plane && editingmode < modify_poly)
		{
			if (editingmode == modify_plane) win(1, 128, 24 + 12 * 8, 255, 35 + 12 * 8);
			if (editingmode != modify_plane) win(0, 128, 24 + 12 * 8, 255, 35 + 12 * 8);
			sprintf(tmptxt, "1 Plane");
			textout(offscreen, font, tmptxt, 132, 24 + 12 * 8 + 2, whitec);
		}
		
		if (editingmode == modify || editingmode >= modify_poly && editingmode < modify_point)
		{
			if (editingmode == modify_poly) win(1, 128, 24 + 12 * 9, 255, 35 + 12 * 9);
			if (editingmode != modify_poly) win(0, 128, 24 + 12 * 9, 255, 35 + 12 * 9);
			sprintf(tmptxt, "2 Poly");
			textout(offscreen, font, tmptxt, 132, 24 + 12 * 9 + 2, whitec);
		}
		
		if (editingmode == modify || editingmode >= modify_point && editingmode < modify_light)
		{
			if (editingmode == modify_point) win(1, 128, 24 + 12 * 10, 255, 35 + 12 * 10);
			if (editingmode != modify_point) win(0, 128, 24 + 12 * 10, 255, 35 + 12 * 10);
			sprintf(tmptxt, "3 Point");
			textout(offscreen, font, tmptxt, 132, 24 + 12 * 10 + 2, whitec);
		}
		
		if (editingmode == modify || editingmode >= modify_light && editingmode < 900)
		{
			if (editingmode == modify_point) win(1, 128, 24 + 12 * 11, 255, 35 + 12 * 11);
			if (editingmode != modify_point) win(0, 128, 24 + 12 * 11, 255, 35 + 12 * 11);
			sprintf(tmptxt, "4 Light");
			textout(offscreen, font, tmptxt, 132, 24 + 12 * 11 + 2, whitec);
		}
		
		if (editingmode == modify && freeze < timer)
		{
			if ((mouse_b & 1 && checkmouse(128, 24 + 12 * 8, 255, 35 + 12 * 8)) || key[KEY_1])
			{
				editingmode = modify_plane;
				freeze = timer + .002 * fpsr;
				hlpoly = FALSE;
				hlpoint = FALSE;
				nselpoint = 0;
				nselpoly = 0;
			}
			if ((mouse_b & 1 && checkmouse(128, 24 + 12 * 9, 255, 35 + 12 * 9)) || key[KEY_2])
			{
				editingmode = modify_poly;
				freeze = timer + .002 * fpsr;
				hlplane = FALSE;
				hlpoint = FALSE;
				nselpoint = 0;
			}			
			if ((mouse_b & 1 && checkmouse(128, 24 + 12 * 10, 255, 35 + 12 * 10)) || key[KEY_3])
			{
				editingmode = modify_point;
				freeze = timer + .002 * fpsr;
				hlplane = FALSE;
				hlpoly = FALSE;
				nselpoly = 0;
			}
			if ((mouse_b & 1 && checkmouse(128, 24 + 12 * 11, 255, 35 + 12 * 11)) || key[KEY_4])
			{
				editingmode = modify_light;
				freeze = timer + .002 * fpsr;
				hlplane = FALSE;
				hlpoly = FALSE;
				nselpoly = 0;
			}
		}

		if (editingmode >= modify_plane && editingmode < modify_poly)
		{
			if (editingmode == modify_plane && planeinfo[selplane].condensed == FALSE || editingmode >= modify_plane_by_input && editingmode < modify_plane_rotate && planeinfo[selplane].condensed == FALSE)
			{
				if (editingmode == modify_plane_by_input) win(1, 256, 24 + 12 * 8, 383, 35 + 12 * 8);
				if (editingmode != modify_plane_by_input) win(0, 256, 24 + 12 * 8, 383, 35 + 12 * 8);
				sprintf(tmptxt, "1 By Input");
				textout(offscreen, font, tmptxt, 260, 24 + 12 * 8 + 2, whitec);
			}

			if (editingmode == modify_plane || editingmode >= modify_plane_rotate && editingmode < modify_plane_translate)
			{
				if (editingmode == modify_plane_rotate) win(1, 256, 24 + 12 * 9, 383, 35 + 12 * 9);
				if (editingmode != modify_plane_rotate) win(0, 256, 24 + 12 * 9, 383, 35 + 12 * 9);
				sprintf(tmptxt, "2 Rotate");
				textout(offscreen, font, tmptxt, 260, 24 + 12 * 9 + 2, whitec);
			}

			if (editingmode == modify_plane || editingmode >= modify_plane_translate && editingmode < modify_plane_reset_base)
			{
				if (editingmode == modify_plane_translate) win(1, 256, 24 + 12 * 10, 383, 35 + 12 * 10);
				if (editingmode != modify_plane_translate) win(0, 256, 24 + 12 * 10, 383, 35 + 12 * 10);
				sprintf(tmptxt, "3 Translate");
				textout(offscreen, font, tmptxt, 260, 24 + 12 * 10 + 2, whitec);
			}

			if (editingmode == modify_plane && planeinfo[selplane].condensed == FALSE || editingmode >= modify_plane_reset_base && editingmode < modify_plane_set_rotate_base && planeinfo[selplane].condensed == FALSE)
			{
				if (editingmode == modify_plane_reset_base) win(1, 256, 24 + 12 * 11, 383, 35 + 12 * 11);
				if (editingmode != modify_plane_reset_base) win(0, 256, 24 + 12 * 11, 383, 35 + 12 * 11);
				sprintf(tmptxt, "4 Reset Base");
				textout(offscreen, font, tmptxt, 260, 24 + 12 * 11 + 2, whitec);
			}

			if (editingmode == modify_plane || editingmode >= modify_plane_set_rotate_base && editingmode < modify_plane_bind)
			{
				if (editingmode == modify_plane_set_rotate_base) win(1, 256, 24 + 12 * 12, 399, 35 + 12 * 12);
				if (editingmode != modify_plane_set_rotate_base) win(0, 256, 24 + 12 * 12, 399, 35 + 12 * 12);
				sprintf(tmptxt, "5 Set Rotate Base");
				textout(offscreen, font, tmptxt, 260, 24 + 12 * 12 + 2, whitec);
			}

			if (editingmode == modify_plane && endplane > 0 || editingmode >= modify_plane_bind && editingmode < modify_plane_clone && endplane > 0)
			{
				if (editingmode == modify_plane_bind) win(1, 256, 24 + 12 * 13, 383, 35 + 12 * 13);
				if (editingmode != modify_plane_bind) win(0, 256, 24 + 12 * 13, 383, 35 + 12 * 13);
				sprintf(tmptxt, "6 Bind");
				textout(offscreen, font, tmptxt, 260, 24 + 12 * 13 + 2, whitec);
			}

			if (editingmode == modify_plane || editingmode >= modify_plane_clone && editingmode < modify_poly)
			{
				if (editingmode == modify_plane_clone) win(1, 256, 24 + 12 * 14, 383, 35 + 12 * 14);
				if (editingmode != modify_plane_clone) win(0, 256, 24 + 12 * 14, 383, 35 + 12 * 14);
				sprintf(tmptxt, "7 Clone");
				textout(offscreen, font, tmptxt, 260, 24 + 12 * 14 + 2, whitec);
			}


			if (editingmode == modify_plane && freeze < timer)
			{
				if ((mouse_b & 1 && checkmouse(256, 24 + 12 * 8, 383, 35 + 12 * 8)) && planeinfo[selplane].condensed == FALSE || key[KEY_1] && planeinfo[selplane].condensed == FALSE) editingmode = modify_plane_by_input;
				if ((mouse_b & 1 && checkmouse(256, 24 + 12 * 9, 383, 35 + 12 * 9)) || key[KEY_2])
				{
					editingmode = modify_plane_rotate;
					condense_rotate(selplane);
				}
				if ((mouse_b & 1 && checkmouse(256, 24 + 12 * 10, 383, 35 + 12 * 10)) || key[KEY_3])
				{
					editingmode = modify_plane_translate;
					freeze = timer + .002 * fpsr;
				}
				if ((mouse_b & 1 && checkmouse(256, 24 + 12 * 11, 383, 35 + 12 * 11)) && planeinfo[selplane].condensed == FALSE || key[KEY_4] && planeinfo[selplane].condensed == FALSE)
				{
					editingmode = modify_plane_reset_base;
					freeze = timer + .002 * fpsr;
					hlpoint = TRUE;
					nselpoint = 0;
				}
				if ((mouse_b & 1 && checkmouse(256, 24 + 12 * 12, 399, 35 + 12 * 12)) || key[KEY_5])
				{
					editingmode = modify_plane_set_rotate_base;
					freeze = timer + .002 * fpsr;
					hlpoint = TRUE;
					nselpoint = 0;
				}
				if ((mouse_b & 1 && checkmouse(256, 24 + 12 * 13, 383, 35 + 12 * 13)) && endplane > 0 || key[KEY_6] && endplane > 0)
				{
					editingmode = modify_plane_bind;
					freeze = timer + .002 * fpsr;
					hlplane = TRUE;					
				}
				if ((mouse_b & 1 && checkmouse(256, 24 + 12 * 14, 383, 35 + 12 * 14)) || key[KEY_7])
				{
					editingmode = modify_plane_clone;
					freeze = timer + .002 * fpsr;
				}
			}

			if (editingmode >= modify_plane_translate && editingmode < modify_plane_reset_base)
			{
				if (editingmode == modify_plane_translate || editingmode == modify_plane_translate_direct)
				{
					if (editingmode == modify_plane_translate_direct) win(1, 384, 24 + 12 * 10, 511, 35 + 12 * 10);
					if (editingmode != modify_plane_translate_direct) win(0, 384, 24 + 12 * 10, 511, 35 + 12 * 10);
					sprintf(tmptxt, "1 Direct");
					textout(offscreen, font, tmptxt, 388, 24 + 12 * 10 + 2, whitec);
				}
				if (editingmode == modify_plane_translate || editingmode == modify_plane_translate_free)
				{
					if (editingmode == modify_plane_translate_free) win(1, 384, 24 + 12 * 11, 511, 35 + 12 * 11);
					if (editingmode != modify_plane_translate_free) win(0, 384, 24 + 12 * 11, 511, 35 + 12 * 11);
					sprintf(tmptxt, "2 Free");
					textout(offscreen, font, tmptxt, 388, 24 + 12 * 11 + 2, whitec);
				}

				if (editingmode == modify_plane_translate && freeze < timer)
				{
					if ((mouse_b & 1 && checkmouse(384, 24 + 12 * 10, 511, 35 + 12 * 10)) || key[KEY_1])
					{
						editingmode = modify_plane_translate_direct;
						if (planeinfo[selplane].condensed) condense_rotate(selplane);
						hlpoint = TRUE;
						nselpoint = 0;
					}
					if ((mouse_b & 1 && checkmouse(384, 24 + 12 * 11, 511, 35 + 12 * 11)) || key[KEY_2])
					{
						editingmode = modify_plane_translate_free;
						if (planeinfo[selplane].condensed) condense_rotate(selplane);
					}
				}
			}
		}

		if (editingmode >= modify_poly && editingmode < modify_point && nselpoly > 0)
		{
			if (editingmode == modify_poly || editingmode >= modify_poly_reset_vertii && editingmode < modify_poly_texture)
			{
				if (editingmode == modify_poly_reset_vertii) win(1, 256, 24 + 12 * 9, 383, 35 + 12 * 9);
				if (editingmode != modify_poly_reset_vertii) win(0, 256, 24 + 12 * 9, 383, 35 + 12 * 9);
				sprintf(tmptxt, "1 Reset Vertii");
				textout(offscreen, font, tmptxt, 260, 24 + 12 * 9 + 2, whitec);
			}

			if (editingmode == modify_poly)
			{
				win(0, 256, 24 + 12 * 10, 383, 35 + 12 * 10);
				sprintf(tmptxt, "2 Texture");
				textout(offscreen, font, tmptxt, 260, 24 + 12 * 10 + 2, whitec);
			}

			if (editingmode == modify_poly)
			{
				win(0, 256, 24 + 12 * 11, 383, 35 + 12 * 11);
				sprintf(tmptxt, "3 UV Map");
				textout(offscreen, font, tmptxt, 260, 24 + 12 * 11 + 2, whitec);
			}

			if (editingmode == modify_poly || editingmode >= modify_poly_hidden && editingmode < modify_poly_clipped)
			{
				if (editingmode == modify_poly_hidden) win(1, 256, 24 + 12 * 12, 383, 35 + 12 * 12);
				if (editingmode != modify_poly_hidden) win(0, 256, 24 + 12 * 12, 383, 35 + 12 * 12);
				sprintf(tmptxt, "4 Hidden On/Off");
				textout(offscreen, font, tmptxt, 260, 24 + 12 * 12 + 2, whitec);
			}
			
			if (editingmode == modify_poly || editingmode >= modify_poly_clipped && editingmode < modify_poly_transperant)
			{
				if (editingmode == modify_poly_clipped) win(1, 256, 24 + 12 * 13, 383, 35 + 12 * 13);
				if (editingmode != modify_poly_clipped) win(0, 256, 24 + 12 * 13, 383, 35 + 12 * 13);
				sprintf(tmptxt, "5 Clipped");
				textout(offscreen, font, tmptxt, 260, 24 + 12 * 13 + 2, whitec);
			}

			if (editingmode == modify_poly || editingmode >= modify_poly_transperant && editingmode < modify_poly_trigger)
			{
				if (editingmode == modify_poly_transperant) win(1, 256, 24 + 12 * 14, 383, 35 + 12 * 14);
				if (editingmode != modify_poly_transperant) win(0, 256, 24 + 12 * 14, 383, 35 + 12 * 14);
				sprintf(tmptxt, "6 Transperant");
				textout(offscreen, font, tmptxt, 260, 24 + 12 * 14 + 2, whitec);
			}

			if (editingmode == modify_poly || editingmode >= modify_poly_trigger && editingmode < 830)
			{
				if (editingmode == modify_poly_trigger) win(1, 256, 24 + 12 * 15, 423, 35 + 12 * 15);
				if (editingmode != modify_poly_trigger) win(0, 256, 24 + 12 * 15, 423, 35 + 12 * 15);
				sprintf(tmptxt, "7 Trigger");
				textout(offscreen, font, tmptxt, 260, 24 + 12 * 15 + 2, whitec);
			}

			if (editingmode == modify_poly && freeze < timer)
			{
				if ((mouse_b & 1 && checkmouse(256, 24 + 12 * 9, 383, 35 + 12 * 9)) || key[KEY_1]) 
				{
					editingmode = modify_poly_reset_vertii;
					nselpoint = 0;
					hlpoint = TRUE;
					freeze = timer + .002 * fpsr;
				}
				if ((mouse_b & 1 && checkmouse(256, 24 + 12 * 10, 383, 35 + 12 * 10)) || key[KEY_2])
				{
					hlpoly = FALSE;
					editingmode = modify_poly_texture;
					freeze = timer + .002 * fpsr;
				}
				if ((mouse_b & 1 && checkmouse(256, 24 + 12 * 11, 383, 35 + 12 * 11)) || key[KEY_3])
				{
					hlpoly = FALSE;
					editingmode = modify_poly_uvmap;
					freeze = timer + .002 * fpsr;
				}
				if ((mouse_b & 1 && checkmouse(256, 24 + 12 * 12, 383, 35 + 12 * 12)) || key[KEY_4])
				{
					for (c = 0; c < nselpoly; c++)
					{
						planeinfo[selplane].polyinfo[selpoly[c]][6] = 1 - planeinfo[selplane].polyinfo[selpoly[c]][6];
					}

					nselpoly = 0;					
					editingmode = 0;
				}	
				if ((mouse_b & 1 && checkmouse(256, 24 + 12 * 13, 383, 35 + 12 * 13)) || key[KEY_5])
				{
					editingmode = 0;
					for (c = 0; c < nselpoly; c++)
					{
						planeinfo[selplane].polyinfo[selpoly[c]][7] = 1 - planeinfo[selplane].polyinfo[selpoly[c]][7];
					}
				}
				if ((mouse_b & 1 && checkmouse(256, 24 + 12 * 14, 383, 35 + 12 * 14)) || key[KEY_6])
				{
					editingmode = modify_poly_transperant;
					freeze = timer + .002 * fpsr;
				}
				if ((mouse_b & 1 && checkmouse(256, 24 + 12 * 15, 423, 35 + 12 * 15)) || key[KEY_7]) editingmode = modify_poly_trigger;
			}
		}

		if (editingmode >= modify_point && editingmode < modify_light && (nselpoint > 0 || nselpointc > 0))
		{
			if (editingmode == modify_point || editingmode >= modify_point_by_input && editingmode < modify_point_rotate)
			{
				if (editingmode == modify_point_by_input) win(1, 256, 24 + 12 * 10, 383, 35 + 12 * 10);
				if (editingmode != modify_point_by_input) win(0, 256, 24 + 12 * 10, 383, 35 + 12 * 10);
				sprintf(tmptxt, "1 By Input");
				textout(offscreen, font, tmptxt, 260, 24 + 12 * 10 + 2, whitec);
			}

			if (editingmode == modify_point || editingmode >= modify_point_rotate && editingmode < modify_point_translate)
			{
				if (editingmode == modify_point_rotate) win(1, 256, 24 + 12 * 11, 383, 35 + 12 * 11);
				if (editingmode != modify_point_rotate) win(0, 256, 24 + 12 * 11, 383, 35 + 12 * 11);
				sprintf(tmptxt, "2 Rotate");
				textout(offscreen, font, tmptxt, 260, 24 + 12 * 11 + 2, whitec);
			}

			if (editingmode == modify_point || editingmode >= modify_point_translate && editingmode < modify_point_clone)
			{
				if (editingmode == modify_point_translate) win(1, 256, 24 + 12 * 12, 383, 35 + 12 * 12);
				if (editingmode != modify_point_translate) win(0, 256, 24 + 12 * 12, 383, 35 + 12 * 12);
				sprintf(tmptxt, "3 Translate");
				textout(offscreen, font, tmptxt, 260, 24 + 12 * 12 + 2, whitec);
			}
						
			if (editingmode == modify_point || editingmode >= modify_point_clone && editingmode < 900)
			{
				if (editingmode == modify_point_clone) win(1, 256, 24 + 12 * 13, 399, 35 + 12 * 13);
				if (editingmode != modify_point_clone) win(0, 256, 24 + 12 * 13, 399, 35 + 12 * 13);
				sprintf(tmptxt, "4 Clone");
				textout(offscreen, font, tmptxt, 260, 24 + 12 * 13 + 2, whitec);
			}


			if (editingmode == modify_point && freeze < timer)
			{
				if ((mouse_b & 1 && checkmouse(256, 24 + 12 * 10, 383, 35 + 12 * 10)) || key[KEY_1]) editingmode = modify_point_by_input;
				if ((mouse_b & 1 && checkmouse(256, 24 + 12 * 11, 383, 35 + 12 * 11)) || key[KEY_2])
				{
					editingmode = modify_point_rotate;
					condense_rotate(selplane);
				}
				if ((mouse_b & 1 && checkmouse(256, 24 + 12 * 12, 383, 35 + 12 * 12)) || key[KEY_3])
				{
					editingmode = modify_point_translate;
					freeze = timer + .002 * fpsr;
				}
				if ((mouse_b & 1 && checkmouse(256, 24 + 12 * 13, 399, 35 + 12 * 13)) || key[KEY_4])
				{
					editingmode = modify_point_clone;
					freeze = timer + .002 * fpsr;
				}
			}

			if (editingmode >= modify_point_translate && editingmode < modify_point_clone)
			{
				if (editingmode == modify_point_translate || editingmode == modify_point_translate_direct)
				{
					if (editingmode == modify_point_translate_direct) win(1, 384, 24 + 12 * 12, 511, 35 + 12 * 12);
					if (editingmode != modify_point_translate_direct) win(0, 384, 24 + 12 * 12, 511, 35 + 12 * 12);
					sprintf(tmptxt, "1 Direct");
					textout(offscreen, font, tmptxt, 388, 24 + 12 * 12 + 2, whitec);
				}
				if (editingmode == modify_point_translate || editingmode == modify_point_translate_free)
				{
					if (editingmode == modify_point_translate_free) win(1, 384, 24 + 12 * 13, 511, 35 + 12 * 13);
					if (editingmode != modify_point_translate_free) win(0, 384, 24 + 12 * 13, 511, 35 + 12 * 13);
					sprintf(tmptxt, "2 Free");
					textout(offscreen, font, tmptxt, 388, 24 + 12 * 13 + 2, whitec);
				}

				if (editingmode == modify_point_translate && freeze < timer)
				{
					if ((mouse_b & 1 && checkmouse(384, 24 + 12 * 12, 511, 35 + 12 * 12)) || key[KEY_1])
					{
						editingmode = modify_point_translate_direct;						
						hlpoint = TRUE;
						
						nselpointc = nselpoint;
						for (c = 0; c < nselpoint; c++)
						{
							selpointc[c][0] = selpoint[c][0];
							selpointc[c][1] = selpoint[c][1];
							selpointc[c][2] = selpoint[c][2];
							selpointc[c][3] = selpoint[c][3];
						}

						nselpoint = 0;

						condense_rotate(selplane);
					}
					if ((mouse_b & 1 && checkmouse(384, 24 + 12 * 13, 511, 35 + 12 * 13)) || key[KEY_2]) 
					{												
						editingmode = modify_point_translate_free;

						condense_rotate(selplane);
					}
				}
			}

		}

		if ((editingmode >= modify_light && editingmode < 900 && nselpoint > 0) || editingmode == modify_light_strength_by_selection)
		{			
			p = selpoint[0][3];
			if (editingmode == modify_light_strength_by_selection) p = selpointc[0][3];
			if (p > -1 || (planeinfo[selplane].lastpoint = 0 && planeinfo[selplane].pointinfo[0][9] > 0))
			{
				if (planeinfo[selplane].lastpoint = 0 && planeinfo[selplane].pointinfo[0][9] > 0) p = 0;
				if (planeinfo[selplane].pointinfo[p][9] > 0)
				{

					if (editingmode == modify_light || editingmode >= modify_light_number && editingmode < modify_light_color)
					{
						if (editingmode == modify_light_number) win(1, 256, 24 + 12 * 11, 383, 35 + 12 * 11);
						if (editingmode != modify_light_number) win(0, 256, 24 + 12 * 11, 383, 35 + 12 * 11);
						sprintf(tmptxt, "1 Light Number");
						textout(offscreen, font, tmptxt, 260, 24 + 12 * 11 + 2, whitec);
					}

					if (editingmode == modify_light || editingmode >= modify_light_color && editingmode < modify_light_strength)
					{
						if (editingmode == modify_light_color) win(1, 256, 24 + 12 * 12, 383, 35 + 12 * 12);
						if (editingmode != modify_light_color) win(0, 256, 24 + 12 * 12, 383, 35 + 12 * 12);
						sprintf(tmptxt, "2 Color");
						textout(offscreen, font, tmptxt, 260, 24 + 12 * 12 + 2, whitec);
					}

					if (editingmode == modify_light || editingmode >= modify_light_strength && editingmode < 900)
					{
						if (editingmode == modify_light_strength) win(1, 256, 24 + 12 * 13, 383, 35 + 12 * 13);
						if (editingmode != modify_light_strength) win(0, 256, 24 + 12 * 13, 383, 35 + 12 * 13);
						sprintf(tmptxt, "3 Strength");
						textout(offscreen, font, tmptxt, 260, 24 + 12 * 13 + 2, whitec);
					}

					if (editingmode == modify_light && freeze < timer)
					{
						if ((mouse_b & 1 && checkmouse(256, 24 + 12 * 11, 383, 35 + 12 * 11)) || key[KEY_1]) editingmode = modify_light_number;
						if ((mouse_b & 1 && checkmouse(256, 24 + 12 * 12, 383, 35 + 12 * 12)) || key[KEY_2]) editingmode = modify_light_color;
						if ((mouse_b & 1 && checkmouse(256, 24 + 12 * 13, 383, 35 + 12 * 13)) || key[KEY_3])
						{
							editingmode = modify_light_strength;
							freeze = timer + .002 * fpsr;
						}
					}

					if (editingmode >= modify_light_strength && editingmode < 900)
					{
						if (editingmode == modify_light_strength || editingmode == modify_light_strength_by_input)
						{
							if (editingmode == modify_light_strength_by_input) win(1, 384, 24 + 12 * 13, 511, 35 + 12 * 13);
							if (editingmode != modify_light_strength_by_input) win(0, 384, 24 + 12 * 13, 511, 35 + 12 * 13);
							sprintf(tmptxt, "1 By Input");
							textout(offscreen, font, tmptxt, 388, 24 + 12 * 13 + 2, whitec);
						}
						if (editingmode == modify_light_strength || editingmode == modify_light_strength_by_selection)
						{
							if (editingmode == modify_light_strength_by_selection) win(1, 384, 24 + 12 * 14, 511, 35 + 12 * 14);
							if (editingmode != modify_light_strength_by_selection) win(0, 384, 24 + 12 * 14, 511, 35 + 12 * 14);
							sprintf(tmptxt, "2 By Selection");
							textout(offscreen, font, tmptxt, 388, 24 + 12 * 14 + 2, whitec);
						}

						if (editingmode == modify_light_strength && freeze < timer)
						{
							if ((mouse_b & 1 && checkmouse(384, 24 + 12 * 13, 511, 35 + 12 * 13)) || key[KEY_1]) editingmode = modify_light_strength_by_input;
							if ((mouse_b & 1 && checkmouse(384, 24 + 12 * 14, 511, 35 + 12 * 14)) || key[KEY_2])
							{
								sellight = planeinfo[selplane].pointinfo[p][9] - 1;
								selpointc[0][0] = selpoint[0][0];
								selpointc[0][1] = selpoint[0][1];
								selpointc[0][2] = selpoint[0][2];
								selpointc[0][3] = selpoint[0][3];
								editingmode = modify_light_strength_by_selection;
								clear_selected();
								hlpoint = TRUE;
							}
							
						}
					}
				}
			}
		}
	}

		

	if (hlplane) win(1, 0, 24 + 12 * 5, 127, 35 + 12 * 5);
	if (!(hlplane)) win(0, 0, 24 + 12 * 5, 127, 35 + 12 * 5);
	sprintf(tmptxt, "F5 Select Plane");
	textout(offscreen, font, tmptxt, 4, 24 + 12 * 5 + 2, whitec);
	if ((mouse_b & 1 && checkmouse(0, 24 + 12 * 5, 127, 35 + 12 * 5))) key[KEY_F5] = -1;
	
	if (hlpoly) win(1, 0, 24 + 12 * 6, 127, 35 + 12 * 6);
	if (!(hlpoly)) win(0, 0, 24 + 12 * 6, 127, 35 + 12 * 6);
	sprintf(tmptxt, "F6 Select Poly");
	textout(offscreen, font, tmptxt, 4, 24 + 12 * 6 + 2, whitec);
	if ((mouse_b & 1 && checkmouse(0, 24 + 12 * 6, 127, 35 + 12 * 6))) key[KEY_F6] = -1;
	
	if (hlpoint) win(1, 0, 24 + 12 * 7, 127, 35 + 12 * 7);
	if (!(hlpoint)) win(0, 0, 24 + 12 * 7, 127, 35 + 12 * 7);
	sprintf(tmptxt, "F7 Select Point");
	textout(offscreen, font, tmptxt, 4, 24 + 12 * 7 + 2, whitec);
	if ((mouse_b & 1 && checkmouse(0, 24 + 12 * 7, 127, 35 + 12 * 7))) key[KEY_F7] = -1;
	
	
	win(0, 0, 24 + 12 * 9, 127, 35 + 12 * 9);
	sprintf(tmptxt, "F9 Grid On/Off");
	textout(offscreen, font, tmptxt, 4, 24 + 12 * 9 + 2, whitec);
	if ((mouse_b & 1 && checkmouse(0, 24 + 12 * 9, 127, 35 + 12 * 9))) key[KEY_F9] = -1;
	
	win(0, 0, 24 + 12 * 10, 127, 35 + 12 * 10);
	sprintf(tmptxt, "F10 Fill Mode");
	textout(offscreen, font, tmptxt, 4, 24 + 12 * 10 + 2, whitec);
	if ((mouse_b & 1 && checkmouse(0, 24 + 12 * 10, 127, 35 + 12 * 10))) key[KEY_F10] = -1;
		
	win(0, 0, 24 + 12 * 11, 127, 35 + 12 * 11);
	sprintf(tmptxt, "F11 Save Screen");
	textout(offscreen, font, tmptxt, 4, 24 + 12 * 11 + 2, whitec);
	if ((mouse_b & 1 && checkmouse(0, 24 + 12 * 11, 127, 35 + 12 * 11))) key[KEY_F11] = -1;
		
	win(0, 0, 24 + 12 * 12, 127, 35 + 12 * 12);
	sprintf(tmptxt, "F12 Exit");
	textout(offscreen, font, tmptxt, 4, 24 + 12 * 12 + 2, whitec);
	if ((mouse_b & 1 && checkmouse(0, 24 + 12 * 12, 127, 35 + 12 * 12))) key[KEY_F12] = -1;	
	
	
	short farleft = xres, fartop = yres, farright = 0, farbottom = 0, fx, fy;
	short clearance, clicked = 0;
	float trans[3];
	short basepoint, basepoly, p1, p2, p3;
	float spanx, spany, spanz;

	clearance = 2 * xres / 320;;
		
	if (pointinfo.size() > 0)
	{
		farleft = xres, fartop = yres, farright = 0, farbottom = 0;
		for (c = 0; c <= planeinfo[selplane].lastpoint; c++)
		{
			if (planeinfo[selplane].pointinfo[c][7] == 0)
			{
				p = short(planeinfo[selplane].pointinfo[c][6]);
				if (pointinfo[p].x < farleft && pointinfo[p].x > -1) farleft = short(pointinfo[p].x);
				if (pointinfo[p].x > farright && pointinfo[p].x < xres) farright = short(pointinfo[p].x);
				if (pointinfo[p].y < fartop && pointinfo[p].y > -1) fartop = short(pointinfo[p].y);
				if (pointinfo[p].y > farbottom && pointinfo[p].y < yres) farbottom = short(pointinfo[p].y);
			}			
		}
		surroundbox(0, farleft, fartop, farright, farbottom);

		nx = planeinfo[selplane].rbx + planeinfo[selplane].basex - focusofs[0];
		ny = planeinfo[selplane].rby + planeinfo[selplane].basey - focusofs[1];
		nz = planeinfo[selplane].rbz + planeinfo[selplane].basez - focusofs[2];
	
		sx = nx * prexx + ny * preyx + nz * prezx - camx;
		sy = nx * prexy + ny * preyy + nz * prezy - camy;
		sz = nx * prexz + ny * preyz + nz * prezz - camz;

		if (sz <= -10)
		{
			fx = short(xcenter + sx / sz * xzoomratio);
			fy = short(ycenter + sy / sz * yzoomratio - scrny);
			if (fx > -1 && fx < xres && fy > -1 && fy < yres) circle (offscreen, fx, fy, clearance, greenc);			
		}		

		if (hlpoint)
		{			
			for (a = 0; a <= endplane; a++)
			{
				if (planeinfo[a].active && editingmode != select_point || (editingmode == select_point && planeinfo[a].active && a == selplane))
				{

					for (c = 0; c <= planeinfo[a].lastpoint; c++)
					{
						p = short(planeinfo[a].pointinfo[c][6]);
						if (checkmouse(pointinfo[p].x-clearance, pointinfo[p].y-clearance, pointinfo[p].x+clearance, pointinfo[p].y+clearance) && clicked == 0 && planeinfo[a].pointinfo[c][7] == 0)
						{
							rect(offscreen, int(pointinfo[p].x) - clearance-1, int(pointinfo[p].y) - clearance-1, int(pointinfo[p].x) + clearance+1, int(pointinfo[p].y) + clearance+1, bluec);
							tselpointinfo[0] = planeinfo[a].pointinfo[c][3] + planeinfo[a].basex;
							tselpointinfo[1] = planeinfo[a].pointinfo[c][4] + planeinfo[a].basey;
							tselpointinfo[2] = planeinfo[a].pointinfo[c][5] + planeinfo[a].basez;
							if (mouse_b & 1 && freeze < timer)
							{						
								
								p = 0;
								if (editingmode == select_point)
								{
									for (b = 0; b < nselpoint; b++)
									{
										if (selpoint[b][3] == -1 && checktolerance(selpoint[b][0], selpoint[b][1], selpoint[b][2], planeinfo[a].pointinfo[c][3] + planeinfo[a].basex, planeinfo[a].pointinfo[c][4] + planeinfo[a].basey, planeinfo[a].pointinfo[c][5] + planeinfo[a].basez))
										{
											selpoint[b][3] = c;
										}
										if (selpoint[b][3] == c)
										{
											p = -1;
										}
									}
								}
								if (p == 0)
								{
									selpoint[nselpoint][0] = planeinfo[a].pointinfo[c][3] + planeinfo[a].basex;
									selpoint[nselpoint][1] = planeinfo[a].pointinfo[c][4] + planeinfo[a].basey;
									selpoint[nselpoint][2] = planeinfo[a].pointinfo[c][5] + planeinfo[a].basez;
									selpoint[nselpoint][3] = -1;
									
									if (a == selplane) selpoint[nselpoint][3] = c;

									if (!key[KEY_SPACE] && nselpoint > 1)
									{
										selpoint[0][0] = selpoint[nselpoint][0];
										selpoint[0][1] = selpoint[nselpoint][1];
										selpoint[0][2] = selpoint[nselpoint][2];
										selpoint[0][3] = selpoint[nselpoint][3];
										nselpoint = 0;
									}

									nselpoint++;
									clicked = 1;
									freeze = timer + .002 * fpsr;

								}
							}
						}
					
					}
				}
			}
			
		}

		if (hlplane)
		{			
			p = getpixel(screenbuffer, mouse_x, mouse_y);
			tselplane = -1;
			for (a = 0; a < nlights; a++)
			{
				b = lights[a][0];
				c = lights[a][1];

				if (checkmouse(pointinfo[b].x - clearance, pointinfo[b].y - clearance, pointinfo[b].x + clearance, pointinfo[b].y + clearance))
				{
					circle(offscreen, pointinfo[b].x, pointinfo[b].y, clearance * 2, bluec);
					tselplane = pointinfo[b].plane;
					
					selpoint[0][0] = planeinfo[tselplane].basex + planeinfo[tselplane].pointinfo[pointinfo[b].point][3];
					selpoint[0][1] = planeinfo[tselplane].basey + planeinfo[tselplane].pointinfo[pointinfo[b].point][4];
					selpoint[0][2] = planeinfo[tselplane].basez + planeinfo[tselplane].pointinfo[pointinfo[b].point][5];
					selpoint[0][3] = pointinfo[b].point;
					nselpoint = 1;
				}
			}			
			if (p < polyinfo.size() || tselplane > -1)
			{
				farleft = xres, fartop = yres, farright = 0, farbottom = 0;
				if (tselplane == -1) tselplane = polyinfo[p].plane;
				for (c = 0; c <= planeinfo[tselplane].lastpoint; c++)
				{

					if (planeinfo[tselplane].pointinfo[c][7] == 0)
					{
						p = planeinfo[tselplane].pointinfo[c][6];
						if (pointinfo[p].x < farleft && pointinfo[p].x > -1) farleft = pointinfo[p].x;
						if (pointinfo[p].x > farright && pointinfo[p].x < xres) farright = pointinfo[p].x;
						if (pointinfo[p].y < fartop && pointinfo[p].y > -1) fartop = pointinfo[p].y;
						if (pointinfo[p].y > farbottom && pointinfo[p].y < yres) farbottom = pointinfo[p].y;
					}

				}
				surroundbox(1, farleft, fartop, farright, farbottom);
				if (mouse_b & 1 && freeze < timer)
				{									
					hlplane = FALSE;					
					if (editingmode == modify_plane_bind && selplane == tselplane)
					{
						editingmode = 0;
					}
					if (editingmode == modify_plane_bind && selplane != tselplane)
					{
						planeinfo[selplane].condensed = TRUE;
						planeinfo[tselplane].active = FALSE;
						
						basepoint = planeinfo[selplane].lastpoint + 1;

						condense_rotate(selplane);
						condense_rotate(tselplane);

						for (c = 0; c <= planeinfo[tselplane].lastpoint; c++)
						{
							if (planeinfo[tselplane].pointinfo[c][7] == 0)
							{
								p = c + basepoint;
								if (p > 999) p = 999;
								spanx = planeinfo[tselplane].basex - planeinfo[selplane].basex;
								spany = planeinfo[tselplane].basey - planeinfo[selplane].basey;
								spanz = planeinfo[tselplane].basez - planeinfo[selplane].basez;
								x = planeinfo[tselplane].pointinfo[c][0] + spanx;
								y = planeinfo[tselplane].pointinfo[c][1] + spany;
								z = planeinfo[tselplane].pointinfo[c][2] + spanz;

								for (a = 0; a <= planeinfo[selplane].lastpoint; a++)
								{
									if (planeinfo[selplane].pointinfo[a][7] == 0 && checktolerance(planeinfo[selplane].pointinfo[a][0], planeinfo[selplane].pointinfo[a][1], planeinfo[selplane].pointinfo[a][2], x, y, z))
									{
										planeinfo[tselplane].pointinfo[c][7] = 1;
										planeinfo[selplane].pointinfo[p][8] = a;
										if (planeinfo[selplane].pointinfo[a][9] == 0 && planeinfo[selplane].pointinfo[p][9] > 0) planeinfo[selplane].pointinfo[a][9] = planeinfo[selplane].pointinfo[p][9];
										break;
									}
								}

								planeinfo[selplane].pointinfo[p][0] = x;
								planeinfo[selplane].pointinfo[p][1] = y;
								planeinfo[selplane].pointinfo[p][2] = z;

								planeinfo[selplane].pointinfo[p][7] = planeinfo[tselplane].pointinfo[c][7];
								planeinfo[selplane].pointinfo[p][9] = planeinfo[tselplane].pointinfo[c][9];
								planeinfo[selplane].pointinfo[p][10] = planeinfo[tselplane].pointinfo[c][10];
							
								if (p > planeinfo[selplane].lastpoint) planeinfo[selplane].lastpoint = p;
							}
						}

						basepoly = planeinfo[selplane].lastpoly + 1;
						for (c = 0; c <= planeinfo[tselplane].lastpoly; c++)
						{
							if (planeinfo[tselplane].polyinfo[c][6] < 2)
							{
								p = c + basepoly;
								if (p > 999) p = 999;
								p1 = planeinfo[tselplane].polyinfo[c][0] + basepoint;
								p2 = planeinfo[tselplane].polyinfo[c][1] + basepoint;
								p3 = planeinfo[tselplane].polyinfo[c][2] + basepoint;

								if (planeinfo[selplane].pointinfo[p1][7] == 1) p1 = planeinfo[selplane].pointinfo[p1][8];
								if (planeinfo[selplane].pointinfo[p2][7] == 1) p2 = planeinfo[selplane].pointinfo[p2][8];
								if (planeinfo[selplane].pointinfo[p3][7] == 1) p3 = planeinfo[selplane].pointinfo[p3][8];


								planeinfo[selplane].polyinfo[p][0] = p1;
								planeinfo[selplane].polyinfo[p][1] = p2;
								planeinfo[selplane].polyinfo[p][2] = p3;
							
								for (i = 3; i < 10; i++)
								{
									planeinfo[selplane].polyinfo[p][i] = planeinfo[tselplane].polyinfo[c][i];
								}
								for (i = 0; i < 6; i++)
								{
									planeinfo[selplane].polyvtx[p][i] = planeinfo[tselplane].polyvtx[c][i];
								}
							
								if (p > planeinfo[selplane].lastpoly) planeinfo[selplane].lastpoly = p;
							}
						}

						rotate_plane(selplane);
						//condense_rotate(selplane);

						tselplane = selplane;
						nplanes--;

						hlplane = TRUE;
						if (nplanes == 1)
						{
							hlplane = FALSE;
							editingmode = 0;
						}

						freeze = timer + .002 * fpsr;
					}
					
					selplane = tselplane;
				}
			}
		}

		if (hlpoly && freeze < timer)
		{
			p = getpixel(screenbuffer, mouse_x, mouse_y);
			if (p < polyinfo.size())
			{
				tselplane = polyinfo[p].plane;
				if (tselplane == selplane)
				{
					tselpoly = polyinfo[p].poly;
					if (mouse_b & 1 && freeze < timer) 
					{
						p = 0;
						for (c = 0; c < nselpoly; c++)
						{
							if (selpoly[c] == tselpoly) p = -1;
						}
						if (p == 0)
						{
							selpoly[nselpoly] = tselpoly;
							nselpoly++;
							if (!key[KEY_SPACE] && nselpoly > 1)
							{							
								selpoly[0] = selpoly[nselpoly-1];
								nselpoly = 1;
							}
						}
					}
				}
			}
		}

	}
	
	for (a = 0; a < nlights; a++)
	{
		p = lights[a][0];
		c = lights[a][1];

		nx = -lightinfo[c].strength;
		ny = 0;
		nz = 0;
	
		sx = nx * prexx + ny * preyx + nz * prezx + pointinfo[p].psx - vdist;
		sy = nx * prexy + ny * preyy + nz * prezy + pointinfo[p].psy - vdist;
		sz = nx * prexz + ny * preyz + nz * prezz + pointinfo[p].psz - vdist;

		if (pointinfo[p].sz > -10) continue;
		fx = xcenter + sx / sz * xzoomratio;
		if (pointinfo[p].x < 0 || pointinfo[p].x > xres) continue;
	
		fy = ycenter + sy / sz * yzoomratio - scrny;
		if (pointinfo[p].y < 0 || pointinfo[p].y > yres) continue;

		x = abs(pointinfo[p].x - fx);
		
		circle (offscreen, pointinfo[p].x, pointinfo[p].y, clearance, orangec);
		circle (offscreen, pointinfo[p].x, pointinfo[p].y, x, makecol(lightinfo[c].lightr, lightinfo[c].lightg, lightinfo[c].lightb));		
	}
	
		
	for (c = 0; c < nselpoint; c++)
	{
		nx = selpoint[c][0] - focusofs[0];
		ny = selpoint[c][1] - focusofs[1];
		nz = selpoint[c][2] - focusofs[2];
	
		sx = nx * prexx + ny * preyx + nz * prezx - camx;
		sy = nx * prexy + ny * preyy + nz * prezy - camy;
		sz = nx * prexz + ny * preyz + nz * prezz - camz;

		if (sz > -10) continue;

		fx = xcenter + sx / sz * xzoomratio;

		fx = xcenter + short(sx / sz * xzoomratio);

		if (fx < 0 || fx > xres) continue;
	

		fy = ycenter + sy / sz * yzoomratio - scrny;

		fy = ycenter + short(sy / sz * yzoomratio) - scrny;

		if (fy < 0 || fy > yres) continue;
		
		rect (offscreen, fx-clearance, fy-clearance, fx+clearance, fy+clearance, whitec);
					
	}

	for (c = 0; c < nselpointc; c++)
	{
		nx = selpointc[c][0] - focusofs[0];
		ny = selpointc[c][1] - focusofs[1];
		nz = selpointc[c][2] - focusofs[2];
	
		sx = nx * prexx + ny * preyx + nz * prezx - camx;
		sy = nx * prexy + ny * preyy + nz * prezy - camy;
		sz = nx * prexz + ny * preyz + nz * prezz - camz;

		if (sz > -10) continue;
		fx = xcenter + sx / sz * xzoomratio;
		if (fx < 0 || fx > xres) continue;
	
		fy = ycenter + sy / sz * yzoomratio - scrny;
		if (fy < 0 || fy > yres) continue;
		
		rect (offscreen, fx-clearance, fy-clearance, fx+clearance, fy+clearance, whitec);
					
	}


	if (editingmode == addlight && nselpoint == 1)
	{

		for (c = 0; c < max_planes; c++)
		{
			if (planeinfo[c].active == FALSE)
			{
				newplane = c;
				break;
			}
		}
		
		planeinfo[newplane].active = TRUE;
		planeinfo[newplane].condensed = TRUE;
		planeinfo[newplane].lastpoint = 0;
		planeinfo[newplane].lastpoly = 0;
		planeinfo[newplane].basex = selpoint[0][0];
		planeinfo[newplane].basey = selpoint[0][1];
		planeinfo[newplane].basez = selpoint[0][2];
		planeinfo[newplane].endx = selpoint[0][0];
		planeinfo[newplane].endy = selpoint[0][1];
		planeinfo[newplane].endz = selpoint[0][2];
		planeinfo[newplane].rbx = 0;
		planeinfo[newplane].rby = 0;
		planeinfo[newplane].rbz = 0;
		planeinfo[newplane].xrot = 0;
		planeinfo[newplane].yrot = 0;
		planeinfo[newplane].zrot = 0;
		planeinfo[newplane].pointinfo[0][9] = lastlight + 1;

		lightinfo[lastlight].strength = 6;
		lightinfo[lastlight].lightr = 255;
		lightinfo[lastlight].lightg = 255;
		lightinfo[lastlight].lightb = 224;		

		if (newplane > endplane) endplane = newplane;

		lastlight++;

		clear_selected();
		editingmode = 0;
		selplane = newplane;

		selpoint[0][0] = planeinfo[selplane].basex;
		selpoint[0][1] = planeinfo[selplane].basey;
		selpoint[0][2] = planeinfo[selplane].basez;
		selpoint[0][3] = 0;
		nselpoint = 1;
	}

	if (editingmode == modify_light_strength_by_selection && nselpoint == 2)
	{
		float dist;

		x = selpoint[1][0] - selpoint[0][0];
		y = selpoint[1][1] - selpoint[0][1];
		z = selpoint[1][2] - selpoint[0][2];

		dist = sqrt(x * x + y * y + z * z);

		lightinfo[sellight].strength = dist;

		nselpoint = 1;
		selpoint[0][0] = selpointc[0][0];
		selpoint[0][1] = selpointc[0][1];
		selpoint[0][2] = selpointc[0][2];
		selpoint[0][3] = selpointc[0][3];
		hlpoint = FALSE;
		editingmode = 0;
	}

	if (editingmode == modify_poly_reset_vertii && nselpoint == 3)
	{
		basepoint = planeinfo[selplane].lastpoint + 1;
		planeinfo[selplane].condensed = TRUE;

		condense_rotate(selplane);
		
		for (c = 0; c < 3; c++)
		{
			p = selpoint[c][3];
			if (p == -1)
			{
				p = c + basepoint;
				selpoint[c][3] = p;
			}

			planeinfo[selplane].pointinfo[p][0] = selpoint[c][0] - planeinfo[selplane].basex;
			planeinfo[selplane].pointinfo[p][1] = selpoint[c][1] - planeinfo[selplane].basey;
			planeinfo[selplane].pointinfo[p][2] = selpoint[c][2] - planeinfo[selplane].basez;

			if (p > planeinfo[selplane].lastpoint) planeinfo[selplane].lastpoint = p;
		}

		basepoly = selpoly[0];
		planeinfo[selplane].polyinfo[basepoly][0] = selpoint[0][3];
		planeinfo[selplane].polyinfo[basepoly][1] = selpoint[1][3];
		planeinfo[selplane].polyinfo[basepoly][2] = selpoint[2][3];
				
		nselpoly = 1;		

		rotate_plane(selplane);
		nselpoint = 0;
		freeze = timer + .002 * fpsr;
	}
	
	if (editingmode == addpoly && nselpoint == 3)
	{
		basepoint = planeinfo[selplane].lastpoint + 1;		

		condense_rotate(selplane);
		
		for (c = 0; c < 3; c++)
		{
			p = selpoint[c][3];
			if (p == -1)
			{
				p = basepoint++;
				selpoint[c][3] = p;
			}

			planeinfo[selplane].pointinfo[p][3] = selpoint[c][0] - planeinfo[selplane].basex;
			planeinfo[selplane].pointinfo[p][4] = selpoint[c][1] - planeinfo[selplane].basey;
			planeinfo[selplane].pointinfo[p][5] = selpoint[c][2] - planeinfo[selplane].basez;

			if (p > planeinfo[selplane].lastpoint) planeinfo[selplane].lastpoint = p;
		}

		basepoly = planeinfo[selplane].lastpoly + 1;
		planeinfo[selplane].polyinfo[basepoly][0] = selpoint[0][3];
		planeinfo[selplane].polyinfo[basepoly][1] = selpoint[1][3];
		planeinfo[selplane].polyinfo[basepoly][2] = selpoint[2][3];
		for (c = 3; c < 10; c++)
		{
			planeinfo[selplane].polyinfo[basepoly][c] = 0;
		}
		for (c = 0; c < 6; c++)
		{
			planeinfo[selplane].polyvtx[basepoly][c] = 0;
		}
		
		planeinfo[selplane].lastpoly = basepoly;

		nselpoly = 1;
		selpoly[0] = basepoly;

		condense_rotate(selplane);
		nselpoint = 0;
		freeze = timer + .002 * fpsr;
	}

	if (editingmode == modify_plane_translate_direct && nselpoint == 2)
	{								
		trans[0] = selpoint[1][0] - selpoint[0][0];
		trans[1] = selpoint[1][1] - selpoint[0][1];
		trans[2] = selpoint[1][2] - selpoint[0][2];

		planeinfo[selplane].basex += trans[0];
		planeinfo[selplane].basey += trans[1];
		planeinfo[selplane].basez += trans[2];
		
		rotate_plane(selplane);
					
		nselpoint = 0;
		hlpoint = FALSE;
		editingmode = 0;
	}

	if (editingmode == modify_plane_reset_base && nselpoint == 2)
	{								
		planeinfo[selplane].basex = selpoint[0][0];
		planeinfo[selplane].basey = selpoint[0][1];
		planeinfo[selplane].basez = selpoint[0][2];
		planeinfo[selplane].endx = selpoint[1][0];
		planeinfo[selplane].endy = selpoint[1][1];
		planeinfo[selplane].endz = selpoint[1][2];
		
		setup_plane(selplane, 0);
					
		nselpoint = 0;
		hlpoint = FALSE;
		editingmode = 0;
	}

	if (editingmode == modify_plane_set_rotate_base && nselpoint == 1)
	{										
		planeinfo[selplane].rbx = selpoint[0][0] - planeinfo[selplane].basex;
		planeinfo[selplane].rby = selpoint[0][1] - planeinfo[selplane].basey;
		planeinfo[selplane].rbz = selpoint[0][2] - planeinfo[selplane].basez;
		
		planeinfo[selplane].condensed = TRUE;
		
		condense_rotate(selplane);
		rotate_plane(selplane);
					
		nselpoint = 0;
		hlpoint = FALSE;
		editingmode = 0;
	}

	
	if (editingmode == modify_plane_clone)
	{		
		for (c = 0; c < max_planes; c++)
		{
			if (planeinfo[c].active == FALSE)
			{
				newplane = c;
				break;
			}
		}
					
		planeinfo[newplane].active = TRUE;
		planeinfo[newplane].basex = planeinfo[selplane].basex;
		planeinfo[newplane].basey = planeinfo[selplane].basey;
		planeinfo[newplane].basez = planeinfo[selplane].basez;
		planeinfo[newplane].endx = planeinfo[selplane].endx;
		planeinfo[newplane].endy = planeinfo[selplane].endy;
		planeinfo[newplane].endz = planeinfo[selplane].endz;
		planeinfo[newplane].xrot = planeinfo[selplane].xrot;
		planeinfo[newplane].yrot = planeinfo[selplane].yrot;
		planeinfo[newplane].zrot = planeinfo[selplane].zrot;
		planeinfo[newplane].rbx = planeinfo[selplane].rbx;
		planeinfo[newplane].rby = planeinfo[selplane].rby;
		planeinfo[newplane].rbz = planeinfo[selplane].rbz;
		planeinfo[newplane].height = planeinfo[selplane].height;
		planeinfo[newplane].rows = planeinfo[selplane].rows;
		planeinfo[newplane].cols = planeinfo[selplane].cols;
		planeinfo[newplane].lastpoint = planeinfo[selplane].lastpoint;
		planeinfo[newplane].lastpoly = planeinfo[selplane].lastpoly;
		planeinfo[newplane].condensed = planeinfo[selplane].condensed;

		for (c = 0; c <= planeinfo[selplane].lastpoint; c++)
		{
			for (b = 0; b < 11; b++)
			{
				planeinfo[newplane].pointinfo[c][b] = planeinfo[selplane].pointinfo[c][b];
			}
		}

		for (c = 0; c <= planeinfo[selplane].lastpoly; c++)
		{			
			for (b = 0; b < 10; b++)
			{
				planeinfo[newplane].polyinfo[c][b] = planeinfo[selplane].polyinfo[c][b];				
			}
			for (b = 0; b < 6; b++)
			{				
				planeinfo[newplane].polyvtx[c][b] = planeinfo[selplane].polyvtx[c][b];
			}
			if (planeinfo[newplane].radiosity_setup[c] == 2) planeinfo[newplane].radiosity_setup[c] = 1;
		}

		if (newplane > endplane) endplane = newplane;
		nplanes++;
		
		rotate_plane(newplane);	
		selplane = newplane;

		editingmode = 0;
	}

	if (editingmode == modify_point_clone)
	{		
		condense_rotate(selplane);

		basepoint = planeinfo[selplane].lastpoint + 1;

		b = 0;
		for (c = 0; c < nselpoint; c++)
		{
			p = -1;
			if (selpoint[c][3] == -1)
			{
				for (a = 0; a <= planeinfo[selplane].lastpoint; a++)
				{
					x = planeinfo[selplane].pointinfo[a][3] + planeinfo[selplane].basex;
					y = planeinfo[selplane].pointinfo[a][4] + planeinfo[selplane].basey;
					z = planeinfo[selplane].pointinfo[a][5] + planeinfo[selplane].basez;
					if (checktolerance(selpoint[c][0], selpoint[c][1], selpoint[c][2], x, y, z))
					{
						p = a;
						break;
					}
				}
			} else {
				p = selpoint[c][3];
			}
		
			if (p > -1)
			{
				planeinfo[selplane].pointinfo[basepoint + b][0] = planeinfo[selplane].pointinfo[p][0];
				planeinfo[selplane].pointinfo[basepoint + b][1] = planeinfo[selplane].pointinfo[p][1];
				planeinfo[selplane].pointinfo[basepoint + b][2] = planeinfo[selplane].pointinfo[p][2];

				if (basepoint + b > planeinfo[selplane].lastpoint) planeinfo[selplane].lastpoint = basepoint + b;
				selpoint[c][3] = basepoint + b;
				
				b++;
			}
		}
	
		rotate_plane(selplane);			
		editingmode = 0;
	}


	if (editingmode == modify_poly_texture && freeze < timer)
	{
		short rettx;

		freeze = timer + .004 * fpsr;
		rettx = gui_select_texture(planeinfo[selplane].polyinfo[selpoly[0]][3]);

		for (c = 0; c < nselpoly; c++)
		{
			planeinfo[selplane].polyinfo[selpoly[c]][3] = rettx;
			blit(textures[rettx], planeinfo[selplane].radiosity[selpoly[c]], 0, 0, 0, 0, 16, 16);
		}

		editingmode = 0;
		setup_scene();
		freeze = timer + .002 * fpsr;

	}

	if (editingmode == modify_poly_uvmap && freeze < timer)
	{
		short top, lft, width, height, a, b, c, d, e, x, y, i, dark = makecol(33, 64, 102), bx, by, w;
		float s, g;
		int whitec = makecol(192, 192, 192);

		a = (nvtx - 1) / 6 + 1;
		c = nvtx;
		if (c > 6) c = 6;
		
		w = 50;
		s = w - (float)w * .1;
		g = (float)s / 30;

		height = a * w + 20;
		width = c * w + 20;

		lft = xcenter - width / 2;
		if (lft < 128) lft = 128;
		top = ycenter - height / 2;

		win(0, lft, top, lft + width - 1 + 6, top + height - 1);
		win(1, lft, top-12, lft + width - 1 + 6 - 7 * 8, top - 1);

		rectfill(offscreen, lft + 6, top + 6, lft + width - 7, top + height - 7, dark);

		sprintf(tmptxt, "Select UV Map");
		textout(offscreen, font, tmptxt, lft + 4, top - 12 + 2, whitec);

		if (checkmouse(lft + width + 6 - 7 * 8, top - 12, lft + width - 1 + 6, top - 1))
		{
			win(1, lft + width + 6 - 7 * 8, top - 12, lft + width - 1 + 6, top - 1);
			if (mouse_b & 1)
			{
				editingmode = 0;				
				setup_scene();
				freeze = timer + .002 * fpsr;
				return;
			}
		} else {
			win(0, lft + width + 6 - 7 * 8, top - 12, lft + width - 1 + 6, top - 1);
		}

		sprintf(tmptxt, "Close");
		textout(offscreen, font, tmptxt, lft + width + 6 - 7 * 8 + 4, top - 12 + 2, whitec);

		for (y = 0; y < a; y++)
		{
			for (x = 0; x < c; x++)
			{
				b = y * c + x;				
				if (b < nvtx)
				{										
					bx = lft + 13 + x * w;
					by = top + 13 + y * w;
					p = planeinfo[selplane].polyinfo[selpoly[0]][3];
					stretch_sprite(offscreen, textures[p], bx, by, s, s);

					line (offscreen, bx + (float)vtx[b][0] / 16 * s + g, by + (float)vtx[b][1] / 16 * s + g, bx + (float)vtx[b][2] / 16 * s + g, by + (float)vtx[b][3] / 16 * s + g, whitec);
					line (offscreen, bx + (float)vtx[b][4] / 16 * s + g, by + (float)vtx[b][5] / 16 * s + g, bx + (float)vtx[b][2] / 16 * s + g, by + (float)vtx[b][3] / 16 * s + g, whitec);
					line (offscreen, bx + (float)vtx[b][0] / 16 * s + g, by + (float)vtx[b][1] / 16 * s + g, bx + (float)vtx[b][4] / 16 * s + g, by + (float)vtx[b][5] / 16 * s + g, whitec);

					d = 0;
					for (e = 0; e < 6; e++)
					{
						if (vtx[b][e] == planeinfo[selplane].polyvtx[selpoly[0]][e]) d++;
					}
					if (d == 6) surroundbox(2, lft + 13 + x * w, top + 13 + y * w, lft + 13 + x * w + s, top + 13 + y * w + s);

					if (checkmouse(lft + 13 + x * w, top + 13 + y * w, lft + 13 + x * w + s, top + 13 + y * w + s))
					{
						surroundbox(3, lft + 13 + x * w, top + 13 + y * w, lft + 13 + x * w + s, top + 13 + y * w + s);
						if (mouse_b & 1)
						{
							for (e = 0; e < nselpoly; e++)
							{
								for (i = 0; i < 6; i++)
								{
									planeinfo[selplane].polyvtx[selpoly[e]][i] = vtx[b][i];
								}
							}
							editingmode = 0;
							setup_scene();
							freeze = timer + .002 * fpsr;
							return;
						}
					}
				}
			}
		}
	}

	if (editingmode == modify_point_translate_direct && nselpoint == 2)
	{										
		trans[0] = selpoint[1][0] - selpoint[0][0];
		trans[1] = selpoint[1][1] - selpoint[0][1];
		trans[2] = selpoint[1][2] - selpoint[0][2];		

		nselpoint = nselpointc;
		for (c = 0; c < nselpointc; c++)
		{
			selpoint[c][0] = selpointc[c][0];
			selpoint[c][1] = selpointc[c][1];
			selpoint[c][2] = selpointc[c][2];
			selpoint[c][3] = selpointc[c][3];
		}

		for (c = 0; c < nselpoint; c++)
		{
			p = selpoint[c][3];
			if (p > -1)
			{
				planeinfo[selplane].pointinfo[p][0] += trans[0];
				planeinfo[selplane].pointinfo[p][1] += trans[1];
				planeinfo[selplane].pointinfo[p][2] += trans[2];				
			}
			selpoint[c][0] += trans[0];
			selpoint[c][1] += trans[1];
			selpoint[c][2] += trans[2];
		}
		
		rotate_plane(selplane);
		editingmode = 0;
		nselpointc = 0;
		hlpoint = FALSE;

	}

	if (editingmode == remove_isolated_points)
	{
		editingmode = 0;
		
		for (i = 0; i <= planeinfo[selplane].lastpoint; i++)
		{
			if (planeinfo[selplane].pointinfo[i][9] == 0) planeinfo[selplane].pointinfo[i][7] = 1;
		}

		for (i = 0; i <= planeinfo[selplane].lastpoly; i++)
		{
			if (planeinfo[selplane].polyinfo[i][6] < 2)
			{
				p1 = planeinfo[selplane].polyinfo[i][0];
				p2 = planeinfo[selplane].polyinfo[i][1];
				p3 = planeinfo[selplane].polyinfo[i][2];

				planeinfo[selplane].pointinfo[p1][7] = 0;
				planeinfo[selplane].pointinfo[p2][7] = 0;
				planeinfo[selplane].pointinfo[p3][7] = 0;
			}
		}		

		freeze = timer + .002 * fpsr;
	}


}


void drawpolys()
{
    int i, poly, p1, p2, p3, t, a, c, x, y, bluec = makecol(0,0,192), redc = makecol(255,0,0), cgrey = makecol(128,128, 128), dbluec = makecol(0, 0, 96), dredc = makecol(128, 0, 0);
	int greyc = makecol(96, 96, 96), whitec = makecol(192,192,192), greenc = makecol(0,192,0), yellowc = makecol(224, 224, 32), orangec = makecol(244, 196, 32);
	short u[3], v[3], cx, cy;
	int trans = makecol(255, 0, 255);
    
    V3D_f v1, v2, v3, vv1, vv2, vv3;

    vispolys = 0;	
	

	float nx, ny, nz, sx, sy, sz, xx, zz;
	short scx[3], scy[3], fx, fy, level;

	short clearance, clicked = 0;
	clearance = 2 * xres / 320;

		
	for (c = 0; c < pointinfo.size(); c++)
	{
		if (pointinfo[c].flag == 0)
		{
			if (planeinfo[pointinfo[c].plane].pointinfo[pointinfo[c].point][7] == 0)
			{
				fx = pointinfo[c].x;
				fy = pointinfo[c].y;
				rectfill(offscreen, fx-1, fy-1, fx+1, fy+1, greyc);				
			}
		}
	}

	if (grid && !generating_radiosity)
	{
		for (xx = -mapx; xx <= mapx; xx+= gridx)
		{
			for (zz = -mapz; zz <= mapz; zz+= gridz)
			{
				nx = xx - focusofs[0];
				ny = gridbase - focusofs[1];
				nz = zz - focusofs[2];				
	
				sx = nx * prexx + ny * preyx + nz * prezx - camx;
				sy = nx * prexy + ny * preyy + nz * prezy - camy;
				sz = nx * prexz + ny * preyz + nz * prezz - camz;

				if (sz > -10) continue;
				fx = short(xcenter + sx / sz * xzoomratio);
				if (fx < 0 || fx > xres) continue;
		
				fy = short(ycenter + sy / sz * yzoomratio - scrny);
				if (fy < 0 || fy > yres) continue;
			
				rectfill(offscreen, fx-1, fy-1, fx+1, fy+1, greyc);
				
				if (hlpoint && editingmode != select_point && clicked == 0)
				{
					if (checkmouse(fx-clearance, fy-clearance, fx+clearance, fy+clearance))
					{
						rect (offscreen, fx-clearance-1, fy-clearance-1, fx+clearance+1, fy+clearance+1, bluec);
						tselpointinfo[0] = xx;
						tselpointinfo[1] = gridbase;
						tselpointinfo[2] = zz;
						if (mouse_b & 1 && freeze < timer && (nselpoint == 0 || (nselpoint >0 && !checktolerance(xx, gridbase, zz, selpoint[nselpoint-1][0], selpoint[nselpoint-1][1], selpoint[nselpoint-1][2]))))
						{
							selpoint[nselpoint][0] = xx;
							selpoint[nselpoint][1] = gridbase;
							selpoint[nselpoint][2] = zz;
							selpoint[nselpoint][3] = -1;
							nselpoint++;
							clicked = 1;
							freeze = timer + .002 * fpsr;

							if (!key[KEY_SPACE])
							{
								selpoint[0][0] = selpoint[nselpoint-1][0];
								selpoint[0][1] = selpoint[nselpoint-1][1];
								selpoint[0][2] = selpoint[nselpoint-1][2];
								nselpoint = 1;
							}
						}
					}
				}
			}
		}
	}
		

	
	for (i = depth; i >= start; i--)
    {
        if (dbuffer[i] == 0)
            continue;

        if (dbuffer[i] > 99)
        {
            poly = short(dbuffer[i] - 100);
    
            p1 = polyinfo[poly].p1;
            p2 = polyinfo[poly].p2;
            p3 = polyinfo[poly].p3;
            t = polyinfo[poly].tx;
			for (a = 0; a < 3; a++)
			{
				u[a] = polyinfo[poly].u[a];
				v[a] = polyinfo[poly].v[a];
			}

            v1.x = pointinfo[p1].x;
            v1.y = pointinfo[p1].y;
			v1.z = pointinfo[p1].sz;
            v1.u = u[0];
            v1.v = v[0];
			v1.c = poly;

            v2.x = pointinfo[p2].x;
            v2.y = pointinfo[p2].y;
			v2.z = pointinfo[p2].sz;
            v2.u = u[1];
            v2.v = v[1];

            v3.x = pointinfo[p3].x;
            v3.y = pointinfo[p3].y;
			v3.z = pointinfo[p3].sz;
            v3.u = u[2];
            v3.v = v[2];
			
			
			
			if (!generating_radiosity)
			{
				if (fillmode == 0)
				{
					c = 0;
					if (hlplane && tselplane == polyinfo[poly].plane) c = bluec;
					if (planeinfo[polyinfo[poly].plane].polyinfo[polyinfo[poly].poly][6] == 1) c = greenc;
					if (polyinfo[poly].plane == selplane) c = yellowc;
					if (polyinfo[poly].plane == selplane && planeinfo[selplane].polyinfo[polyinfo[poly].poly][7] == 1) c = orangec;					
					if (planeinfo[polyinfo[poly].plane].polyinfo[polyinfo[poly].poly][9] == 3 || planeinfo[polyinfo[poly].plane].polyinfo[polyinfo[poly].poly][9] == 4) c = dredc;
					if (planeinfo[polyinfo[poly].plane].polyinfo[polyinfo[poly].poly][9] == 5 || planeinfo[polyinfo[poly].plane].polyinfo[polyinfo[poly].poly][9] == 6) c = dredc;
					if (planeinfo[polyinfo[poly].plane].polyinfo[polyinfo[poly].poly][9] == 7) c = dbluec;
					if (hlpoly && tselpoly == polyinfo[poly].poly && selplane == polyinfo[poly].plane) c = bluec;
					if (planeinfo[polyinfo[poly].plane].polyinfo[polyinfo[poly].poly][9] > 2)
					{
						cx = float(v1.x + v2.x + v3.x) / 3;
						cy = float(v1.y + v2.y + v3.y) / 3;

						line(offscreen, int(v1.x), int(v1.y), cx, cy, c);
						line(offscreen, int(v2.x), int(v2.y), cx, cy, c);
						line(offscreen, int(v3.x), int(v3.y), cx, cy, c);
					}

					line(offscreen, int(v1.x), int(v1.y), int(v2.x), int(v2.y), c);
					line(offscreen, int(v3.x), int(v3.y), int(v2.x), int(v2.y), c);
					line(offscreen, int(v1.x), int(v1.y), int(v3.x), int(v3.y), c);
				}				
				
				if (hlplane || (hlpoly && selplane == polyinfo[poly].plane)) triangle3d_f(screenbuffer, POLYTYPE_FLAT, textures[t], &v1, &v2, &v3);

				if (fillmode == 1 && planeinfo[polyinfo[poly].plane].polyinfo[polyinfo[poly].poly][6] == 0)
				{
					level = planeinfo[polyinfo[poly].plane].polyinfo[polyinfo[poly].poly][8];
					if (level > 0)
					{
						set_trans_blender(255,255,255,255 * (float(100 - level) / 100));
						triangle3d_f(offscreen, POLYTYPE_PTEX_MASK_TRANS, planeinfo[polyinfo[poly].plane].radiosity[polyinfo[poly].poly], &v1, &v2, &v3);
					} else {
						triangle3d_f(offscreen, POLYTYPE_PTEX_MASK, planeinfo[polyinfo[poly].plane].radiosity[polyinfo[poly].poly], &v1, &v2, &v3);	
					}
					
				}
				
				for (c = 0; c < nselpoly; c++)
				{
					if (polyinfo[poly].poly == selpoly[c] && polyinfo[poly].plane == selplane)
					{					
						line(offscreen, int(v1.x), int(v1.y), int(v2.x), int(v2.y), whitec);
						line(offscreen, int(v3.x), int(v3.y), int(v2.x), int(v2.y), whitec);
						line(offscreen, int(v1.x), int(v1.y), int(v3.x), int(v3.y), whitec);
					}
				}
			} else {
				if (planeinfo[polyinfo[poly].plane].polyinfo[polyinfo[poly].poly][6] == 0 && planeinfo[polyinfo[poly].plane].polyinfo[polyinfo[poly].poly][8] == 0)
				{
					triangle3d_f(offscreen, POLYTYPE_PTEX_MASK, planeinfo[polyinfo[poly].plane].radiosity[polyinfo[poly].poly], &v1, &v2, &v3);

					clear_to_color(texturesb, trans);
					blit(texturesc, texturesb, 0, 0, 0, 0, 16, 16);					
					for (x = 0; x < 16; x++)
					{
						for (y = 0; y < 16; y++)
						{
							c = getpixel(textures[t], x, y);
							if (c == trans) putpixel(texturesb, x, y, trans);
						}
					}
					triangle3d_f(d3buffer[1], POLYTYPE_PTEX_MASK, texturesb, &v1, &v2, &v3);
					

					clear_to_color(texturesb, trans);
					for (x = 0; x < 16; x++)
					{
						for (y = 0; y < 16; y++)
						{
							c = getpixel(textures[t], x, y);
							if (c != trans) putpixel(texturesb, x, y, poly);
						}
					}
					triangle3d_f(d3buffer[0], POLYTYPE_PTEX_MASK, texturesb, &v1, &v2, &v3);

					vv1.z = pointinfo[p1].sz;
					vv2.z = pointinfo[p2].sz;
					vv3.z = pointinfo[p3].sz;

														
					clear_to_color(texturesb, trans);					

					vv1.c = (short)pointinfo[p1].psx;
					vv2.c = (short)pointinfo[p2].psx;
					vv3.c = (short)pointinfo[p3].psx;

					vv1.x = u[0];
					vv1.y = v[0];
					vv2.x = u[1];
					vv2.y = v[1];
					vv3.x = u[2];
					vv3.y = v[2];
					triangle3d_f(texturesb, POLYTYPE_GRGB, texturesb, &vv1, &vv2, &vv3);
					vv1.x = 15 - u[0];
					vv1.y = 15 - v[0];
					vv2.x = u[2];
					vv2.y = v[2];
					vv3.x = u[1];
					vv3.y = v[1];
					vv2.c = (short)pointinfo[p3].psx;
					vv3.c = (short)pointinfo[p2].psx;
					triangle3d_f(texturesb, POLYTYPE_GRGB, texturesb, &vv1, &vv2, &vv3);

					for (x = 0; x < 16; x++)
					{
						for (y = 0; y < 16; y++)
						{
							c = getpixel(textures[t], x, y);
							if (c == trans) putpixel(texturesb, x, y, trans);
						}
					}
					triangle3d_f(d3buffer[2], POLYTYPE_PTEX_MASK, texturesb, &v1, &v2, &v3);					

					clear_to_color(texturesb, trans);

					vv1.c = (short)pointinfo[p1].psy;
					vv2.c = (short)pointinfo[p2].psy;
					vv3.c = (short)pointinfo[p3].psy;

					vv1.x = u[0];
					vv1.y = v[0];
					vv2.x = u[1];
					vv2.y = v[1];
					vv3.x = u[2];
					vv3.y = v[2];
					triangle3d_f(texturesb, POLYTYPE_GRGB, texturesb, &vv1, &vv2, &vv3);
					vv1.x = 15 - u[0];
					vv1.y = 15 - v[0];
					vv2.x = u[2];
					vv2.y = v[2];
					vv3.x = u[1];
					vv3.y = v[1];
					vv2.c = (short)pointinfo[p3].psy;
					vv3.c = (short)pointinfo[p2].psy;
					triangle3d_f(texturesb, POLYTYPE_GRGB, texturesb, &vv1, &vv2, &vv3);
					for (x = 0; x < 16; x++)
					{
						for (y = 0; y < 16; y++)
						{
							c = getpixel(textures[t], x, y);
							if (c == trans) putpixel(texturesb, x, y, trans);
						}
					}
					triangle3d_f(d3buffer[3], POLYTYPE_PTEX_MASK, texturesb, &v1, &v2, &v3);

					clear_to_color(texturesb, trans);

					vv1.c = (short)pointinfo[p1].psz;
					vv2.c = (short)pointinfo[p2].psz;
					vv3.c = (short)pointinfo[p3].psz;

					vv1.x = u[0];
					vv1.y = v[0];
					vv2.x = u[1];
					vv2.y = v[1];
					vv3.x = u[2];
					vv3.y = v[2];
					triangle3d_f(texturesb, POLYTYPE_GRGB, texturesb, &vv1, &vv2, &vv3);
					vv1.x = 15 - u[0];
					vv1.y = 15 - v[0];
					vv2.x = u[2];
					vv2.y = v[2];
					vv3.x = u[1];
					vv3.y = v[1];
					vv2.c = (short)pointinfo[p3].psz;
					vv3.c = (short)pointinfo[p2].psz;
					triangle3d_f(texturesb, POLYTYPE_GRGB, texturesb, &vv1, &vv2, &vv3);
					for (x = 0; x < 16; x++)
					{
						for (y = 0; y < 16; y++)
						{
							c = getpixel(textures[t], x, y);
							if (c == trans) putpixel(texturesb, x, y, trans);
						}
					}
					triangle3d_f(d3buffer[4], POLYTYPE_PTEX_MASK, texturesb, &v1, &v2, &v3);					

					
				}
			}
			

			++vispolys;
        }

        if (dbuffer[i] > 1 && dbuffer[i] < 100)
        {
			drawsprite(dbuffer[i]-2);
            /*npc = short(dbuffer[i]) - 2;
  
            ratio = npcinfo[npc][5];
    
			// WARNING: The sprites array and cf have not been initialized at this point!
            stretch_sprite(offscreen, sprites[cf], int(npcinfo[npc][3]), int(npcinfo[npc][4]), int(32 * ratio), int(32 * ratio));
			*/
        }

        if (dbuffer[i] == 1)
        {
			drawsprite(-1);
        }
        dbuffer[i] = 0;
    }


	nx = fightfocus[0] - focusofs[0];
	ny = fightfocus[1] - focusofs[1];
	nz = fightfocus[2] - focusofs[2];
	
	sx = nx * prexx + ny * preyx + nz * prezx - camx;
    sy = nx * prexy + ny * preyy + nz * prezy - camy;
    sz = nx * prexz + ny * preyz + nz * prezz - camz;

	scx[0] = short(getx(sx,sz));
	scy[0] = short(gety(sy,sz));

	nx = fightfocus[0] - focusofs[0] + 1;
	ny = fightfocus[1] - focusofs[1];
	nz = fightfocus[2] - focusofs[2];
	
	sx = nx * prexx + ny * preyx + nz * prezx - camx;
    sy = nx * prexy + ny * preyy + nz * prezy - camy;
    sz = nx * prexz + ny * preyz + nz * prezz - camz;

	scx[1] = short(getx(sx,sz));
	scy[1] = short(gety(sy,sz));

	nx = fightfocus[0] - focusofs[0] + 1;
	ny = fightfocus[1] - focusofs[1];
	nz = fightfocus[2] - focusofs[2] + 1;
	
	sx = nx * prexx + ny * preyx + nz * prezx - camx;
    sy = nx * prexy + ny * preyy + nz * prezy - camy;
    sz = nx * prexz + ny * preyz + nz * prezz - camz;

	scx[2] = short(getx(sx,sz));
	scy[2] = short(gety(sy,sz));

	nx = fightfocus[0] - focusofs[0];
	ny = fightfocus[1] - focusofs[1];
	nz = fightfocus[2] - focusofs[2] + 1;
	
	sx = nx * prexx + ny * preyx + nz * prezx - camx;
    sy = nx * prexy + ny * preyy + nz * prezy - camy;
    sz = nx * prexz + ny * preyz + nz * prezz - camz;

	scx[3] = short(getx(sx,sz));
	scy[3] = short(gety(sy,sz));

	line(offscreen, scx[0], scy[0], scx[1], scy[1], redc);
	line(offscreen, scx[1], scy[1], scx[2], scy[2], redc);
	line(offscreen, scx[2], scy[2], scx[3], scy[3], redc);
	line(offscreen, scx[3], scy[3], scx[0], scy[0], redc);

}

void drawsprite(short spritenum)
{
	if (spritenum == -1) 
	{
		stretch_sprite(offscreen, temp2, pscx, pscy, int(32 * pratio), int(32 * pratio));
	}

}

void generate_radiosity()
{

	short x, y, i, an, j, pl, po, plane, poly, nx, ny, p, or, og, ob, nr, ng, nb, lightr, lightg, lightb, p1, p2, p3;
	short ambientr, ambientg, ambientb, spanr, spang, spanb, focusstep, l, a, b;
	float xr, yr, zr, dist, intense, strength, xx, yy, zz;
	float v, sx, sy, sz, aa, bb, cc, dd, ee, ff, c, c1, c2, c3;
	float ax, ay, az, bx, by, bz, ii, jj, kk, d, d1, d2, normalx, normaly, normalz;
		
	int	cgrey = makecol(128,128,128), whitec = makecol(192,192,192);
	char tmp[100];

	ambientr = lightinfo[0].lightr * .1;
	ambientg = lightinfo[0].lightg * .1;
	ambientb = lightinfo[0].lightb * .1;

	if (nlights == 0)
	{
			camx = bk_camx;
			camy = bk_camy;
			camz = bk_camz;
			camrx = bk_camrx;
			camry = bk_camry;
			camrz = bk_camrz;
			scrny = bk_scrny;
			gridbase = bk_gridbase;
			xzoomratio = bk_xzoomratio;
			yzoomratio = bk_yzoomratio;
			bfc = bk_bfc;

			fightfocus[1] = 0;
			focusofs[1] = 0;

			stp = 8;

			generating_radiosity = FALSE;

			return;
	}

			for (i = 0; i <= endplane; i++)
			{
				condense_rotate(i);				
			}			

			for (l = 0; l < nlights; l++)
			{
				p = lights[l][1];
				c = lights[l][0];
								
				strength = lightinfo[p].strength * 1.33 * stp;
				lightr = lightinfo[p].lightr;
				lightg = lightinfo[p].lightg;
				lightb = lightinfo[p].lightb;

				lightres = yres;

				xzoomratio = float(lightres / 2.6);
				yzoomratio = float(lightres / 2.6);


				for (i = 0; i < 5; i++)
				{
					d3buffer[i] = create_system_bitmap(lightres, lightres);
					clear_to_color(d3buffer[i], cgrey);
				}

				fightfocus[0] = planeinfo[pointinfo[c].plane].pointinfo[pointinfo[c].point][0] + planeinfo[pointinfo[c].plane].basex;
				fightfocus[1] = planeinfo[pointinfo[c].plane].pointinfo[pointinfo[c].point][1] + planeinfo[pointinfo[c].plane].basey;
				fightfocus[2] = planeinfo[pointinfo[c].plane].pointinfo[pointinfo[c].point][2] + planeinfo[pointinfo[c].plane].basez;
				focusofs[0] = fightfocus[0];
				focusofs[1] = fightfocus[1];
				focusofs[2] = fightfocus[2];
		

				for (a = 0; a <= endplane; a++)
				{
					if (planeinfo[a].active)
					{
						for (b = 0; b <= planeinfo[a].lastpoly; b++)
						{
							if (planeinfo[a].polyinfo[b][6] == 0)
							{
											
								for (x = 0; x < 16; x++)
								{
									for (y = 0; y < 16; y++)
									{
										planeinfo[a].radiosity_index[b][x][y] = FALSE;
									}
								}
							}
						}
					}
				}

			
			for (an = 0; an < 6; an++)
			{
				if (an == 0)
				{
					xr = 0;
					yr = 0;
					zr = 0;
				}
				if (an == 1)
				{
					xr = 90;
					yr = 0;
					zr = 0;
				}
				if (an == 2)
				{
					xr = 180;
					yr = 0;
					zr = 0;
				}
				if (an == 3)
				{
					xr = 270;
					yr = 0;
					zr = 0;
				}
				if (an == 4)
				{
					xr = 0;
					yr = 90;
					zr = 0;
				}
				if (an == 5)
				{
					xr = 0;
					yr = 270;
					zr = 0;
				}

						camrx = xr;
						camry = yr;
						camrz = zr;			
						
						clear_to_color(offscreen, cgrey);	
						clear_to_color(d3buffer[0], cgrey);						
						clear_to_color(d3buffer[1], cgrey);

						calcsteps();
						setup_scene();
						calcpoints();
						sortpolys();

						for (i = 0; i < polyinfo.size(); i++)
						{
							p1 = polyinfo[i].p1;
							p2 = polyinfo[i].p2;
							p3 = polyinfo[i].p3;

							sx = pointinfo[p1].psx - vdist;
							sy = pointinfo[p1].psy - vdist;
							sz = pointinfo[p1].psz - vdist;

							ax = pointinfo[p2].psx - vdist - sx;
							ay = pointinfo[p2].psy - vdist - sy;
							az = pointinfo[p2].psz - vdist - sz;

							bx = pointinfo[p3].psx - vdist - sx;
							by = pointinfo[p3].psy - vdist - sy;
							bz = pointinfo[p3].psz - vdist - sz;

							ii = ay * bz - by * az;
							jj = az * bx - bz * ax;
							kk = ax * by - bx * ay;

							d = sqrt(ii * ii + jj * jj + kk * kk);

							polyinfo[i].nx = -ii / d;
							polyinfo[i].ny = -jj / d;
							polyinfo[i].nz = -kk / d;
							

						}
						drawpolys();

						sprintf(tmp, "%d/%d, %f, %f, %f", l, nlights, xr, yr, zr);
						textout(offscreen, font, tmp, 0, 0, 0);

						focusstep = 1;
					
						for (x = 0; x < lightres; x++)
						{
							for (y = 0; y < lightres; y++)
							{
								if (getpixel(d3buffer[0], x, y) < cgrey)
								{
									pl = getpixel(d3buffer[0], x, y);
									po = getpixel(d3buffer[1], x, y);

									plane = polyinfo[pl].plane;
									poly = polyinfo[pl].poly;

									ny = short(po / 16);
									nx = po - short(po / 16) * 16;
									
									if (!planeinfo[plane].radiosity_index[poly][nx][ny])
									{
																					
											
										planeinfo[plane].radiosity_index[poly][nx][ny] = TRUE;
										xx = getpixel(d3buffer[2], x, y) - vdist;
										yy = getpixel(d3buffer[3], x, y) - vdist;
										zz = getpixel(d3buffer[4], x, y) - vdist;

										dist = sqrt(xx * xx + yy * yy + zz * zz);

										v = 1;											
											
											
										c = getpixel(textures[planeinfo[plane].polyinfo[poly][3]], (short)nx, (short)ny);
										nr = getr(c);
										ng = getg(c);
										nb = getb(c);
										//c1 = sin(1.57 * float(nr + ng + nb) / 768) * 256;											
										c1 = float(nr + ng + nb) / 768;
										
										sx = nx;
										sy = ny;
										sx = nx + float(planeinfo[plane].polyvtx[poly][2] - planeinfo[plane].polyvtx[poly][0]) / 15;
										sy = ny + float(planeinfo[plane].polyvtx[poly][3] - planeinfo[plane].polyvtx[poly][1]) / 15;
											
										if (sx < 0) sx = 0;
										if (sx > 15) sx = 15;
										if (sy < 0) sy = 0;
										if (sy > 15) sy = 15;

											
										c = getpixel(textures[planeinfo[plane].polyinfo[poly][3]], (short)sx, (short)sy);
										nr = getr(c);
										ng = getg(c);
										nb = getb(c);
										c2 = float(nr + ng + nb) / 768 - c1;
											
										

										sx = nx;
										sy = ny;
										sx = nx + float(planeinfo[plane].polyvtx[poly][4] - planeinfo[plane].polyvtx[poly][0]) / 15;
										sy = ny + float(planeinfo[plane].polyvtx[poly][5] - planeinfo[plane].polyvtx[poly][1]) / 15;
											
										if (sx < 0) sx = 0;
										if (sx > 15) sx = 15;
										if (sy < 0) sy = 0;
										if (sy > 15) sy = 15;

											
										c = getpixel(textures[planeinfo[plane].polyinfo[poly][3]], (short)sx, (short)sy);
										nr = getr(c);
										ng = getg(c);
										nb = getb(c);
										c3 = float(nr + ng + nb) / 768 - c1;
											
										p1 = polyinfo[pl].p1;
										p2 = polyinfo[pl].p2;
										p3 = polyinfo[pl].p3;
											
										sx = pointinfo[p1].psx - vdist;
										sy = pointinfo[p1].psy - vdist;
										sz = pointinfo[p1].psz - vdist;

										ax = pointinfo[p2].psx - vdist - sx + polyinfo[pl].nx * c2 * 6;
										ay = pointinfo[p2].psy - vdist - sy + polyinfo[pl].ny * c2 * 6;
										az = pointinfo[p2].psz - vdist - sz + polyinfo[pl].nz * c2 * 6;

										bx = pointinfo[p3].psx - vdist - sx + polyinfo[pl].nx * c3 * 6;
										by = pointinfo[p3].psy - vdist - sy + polyinfo[pl].ny * c3 * 6;
										bz = pointinfo[p3].psz - vdist - sz + polyinfo[pl].nz * c3 * 6;
										
										
										ii = ay * bz - by * az;
										jj = az * bx - bz * ax;
										kk = ax * by - bx * ay;

										d = sqrt(ii * ii + jj * jj + kk * kk);

										normalx = -ii / d;
										normaly = -jj / d;
										normalz = -kk / d;
											
										sx = xx;
										sy = yy;
										sz = zz;

										aa = normalx;
										bb = normaly;
										cc = normalz;

										//aa = polyinfo[pl].nx;
										//bb = polyinfo[pl].ny;
										//cc = polyinfo[pl].nz;
											
										dd = -sx;
										ee = -sy;
										ff = -sz;

										d1 = sqrt(aa * aa + bb * bb + cc * cc);
										d2 = sqrt(dd * dd + ee * ee + ff * ff);

										aa /= d1;
										bb /= d1;
										cc /= d1;

										dd /= d2;
										ee /= d2;
										ff /= d2;

										v = aa * dd + bb * ee + cc * ff;

											
										if (v < 0) v = 0;
										if (v > 1) v = 1;
											
										
																																								
										
										if (dist < 0) dist = 0;
										if (dist > strength) dist = strength;
										
										intense = v * (strength - dist) / strength;										

										
																					
										if (intense > 1) intense = 1;
										if (intense < 0) intense = 0;
											
										c = getpixel(planeinfo[plane].radiosity[poly], nx, ny);
										nr = lightr * intense;
										ng = lightg * intense;
										nb = lightb * intense;
										nr = nr + getr(c);
										ng = ng + getg(c);
										nb = nb + getb(c);
										


										if (nr > 255) nr = 255;
										if (ng > 255) ng = 255;
										if (nb > 255) nb = 255;

										putpixel(planeinfo[plane].radiosity[poly], nx, ny, makecol(nr, ng, nb));

										//planeinfo[plane].radiosity_total[poly]++;
										
										poll_keyboard();
										if (key[KEY_ESC]) return;
									}
									
								}
							}						
						}
						
						
						blit(offscreen, screen, 0, 0, 0, 0, lightres, lightres);
						clear_to_color(offscreen, cgrey);							
			
				}

				for (i = 0; i < 5; i++)
				{
					destroy_bitmap(d3buffer[i]);
				}

												
			}
			
			for (ii = 0; ii < 2; ii++)
			{
			for (i = 0; i <= endplane; i++)
			{
				if (planeinfo[i].active)
				{
					for (j = 0; j <= planeinfo[i].lastpoly; j++)
					{
						if (planeinfo[i].polyinfo[j][6] == 0)
						{							
							for (x = 0; x < 16; x++)
							{
								for (y = 0; y < 16; y++)
								{
									nr = 0;
									ng = 0;
									nb = 0;

									c = getpixel(planeinfo[i].radiosity[j], x, y);
									or = getr(c);
									og = getg(c);
									ob = getb(c);

									p = 0;
									for (xx = -1; xx < 2; xx++)
									{
										for (yy = -1; yy < 2; yy++)
										{
											sx = x + xx;
											sy = y + yy;

											if (sx > -1 && sx < 16 && sy > -1 && sy < 16)
											{
												p++;
												c = getpixel(planeinfo[i].radiosity[j], sx, sy);
												nr += getr(c);
												ng += getg(c);
												nb += getb(c);
											}
										}
									}

									if (p > 0)
									{
										nr /= p;
										ng /= p;
										nb /= p;

										if (nr > 255) nr = 255;
										if (ng > 255) ng = 255;
										if (nb > 255) nb = 255;

										if (nr < or) nr = or;
										if (ng < og) ng = og;
										if (nb < ob) nb = ob;
										
										putpixel(planeinfo[i].radiosity[j], x, y, makecol(nr, ng, nb));
									}

								}
							}
						}
					}
																		
				}
			}
			}


			for (i = 0; i <= endplane; i++)
			{
				if (planeinfo[i].active)
				{
					for (j = 0; j <= planeinfo[i].lastpoly; j++)
					{
						if (planeinfo[i].polyinfo[j][6] == 0)
						{
							for (x = 0; x < 16; x++)
							{
								for (y = 0; y < 16; y++)
								{
									c = getpixel(planeinfo[i].radiosity[j], x, y);
									nr = getr(c);
									ng = getg(c);
									nb = getb(c);

									spanr = 255 - ambientr;
									spang = 255 - ambientg;
									spanb = 255 - ambientb;

									nr = ambientr + spanr * (float)nr / 255;
									ng = ambientg + spang * (float)ng / 255;
									nb = ambientb + spanb * (float)nb / 255;
							
									c = getpixel(textures[planeinfo[i].polyinfo[j][3]], x, y);
									nr = getr(c) * (float)nr / 255;
									ng = getg(c) * (float)ng / 255;
									nb = getb(c) * (float)nb / 255;
	
									if (getr(c) == 255 && getg(c) == 0 && getb(c) == 255)
									{
										nr = 255;
										ng = 0;
										nb = 255;
									}
									putpixel(planeinfo[i].radiosity[j], x, y, makecol(nr, ng, nb));
								}
							}
						}
					}
																		
				}				

			}

			camx = bk_camx;
			camy = bk_camy;
			camz = bk_camz;
			camrx = bk_camrx;
			camry = bk_camry;
			camrz = bk_camrz;
			scrny = bk_scrny;
			gridbase = bk_gridbase;
			xzoomratio = bk_xzoomratio;
			yzoomratio = bk_yzoomratio;
			bfc = bk_bfc;

			fightfocus[1] = 0;
			focusofs[1] = 0;

			stp = 8;

			generating_radiosity = FALSE;

}

void mapeditinit()
{
	int ret;

	focusofs[0] = 0;
	focusofs[1] = 0;
	focusofs[2] = 0;

	endplane = 0;
	fillmode = 0;
	bfc = FALSE;
	nselpoint = 0;

	gridx = 1;
	gridz = 1;
	gridy = .5;

	gridbase = 0;
	grid = TRUE;

	lastlight = 0;

	mapx = 50;
	mapz = 50;
	
	camz = 126;		
	camrx = 54;

	hud = TRUE;

	ret = install_mouse();
	
	screenbuffer = create_system_bitmap(xres, yres);

	BITMAP* mousepcx;
	BITMAP* temppcx;
	mousepcx = create_system_bitmap(32, 32);
	temppcx = create_system_bitmap(32, 32);
		
	temppcx = load_pcx("data/mouse.pcx", pal);
	blit(temppcx, mousepcx, 0, 0, 0, 0, 32, 32);

	set_mouse_sprite(mousepcx);
	set_mouse_sprite_focus(2,2);
	rectfill(offscreen, 0, 0, SCREEN_W, SCREEN_H, makecol(128,128,128));

	tolerance = .1;

	char textemp[100];

	temppcx = create_system_bitmap(320, 240);
	sprintf(textemp, "data/textureset/tx0.pcx");
	temppcx = load_pcx(textemp, pal);
    
	
	int x, y, i, a, b, c, xx, yy;
	float sx, sy;
	
	i = 0;
    for (y = 0; y < 15; y++)
    {
        for (x = 0; x < 20; x++)
        {
			i = y * 20 + x;
            blit(temppcx, textures[i], x * 16, y * 16, 0, 0, 16, 16);            
        }
    }

	
	xx = 1;
	yy = 1;
	i = 0;

	sx = 15 / (float)xx;
	sy = 15 / (float)yy;

	for (y = 0; y < yy; y++)
	{
		for (x = 0; x < xx; x++)
		{
			for (a = y + 1; a <= yy; a++)
			{
				for (b = x + 1; b <= xx; b++)
				{
					vtx[i][0] = x * sx;
					vtx[i][1] = y * sy;
					vtx[i][2] = x * sx;
					vtx[i][3] = a * sy;
					vtx[i][4] = b * sx;
					vtx[i][5] = y * sy;

					i++;

					vtx[i][0] = b * sx;
					vtx[i][1] = a * sy;
					vtx[i][2] = b * sx;
					vtx[i][3] = y * sy;
					vtx[i][4] = x * sx;
					vtx[i][5] = a * sy;

					i++;
				}
			}
		}
	}

	for (y = 0; y < yy; y++)
	{
		for (x = 0; x < xx; x++)
		{
			for (a = y + 1; a <= yy; a++)
			{
				for (b = x + 1; b <= xx; b++)
				{
					vtx[i][0] = x * sx;
					vtx[i][1] = a * sy;
					vtx[i][2] = b * sx;
					vtx[i][3] = a * sy;
					vtx[i][4] = x * sx;
					vtx[i][5] = y * sy;

					i++;

					vtx[i][0] = b * sx;
					vtx[i][1] = y * sy;
					vtx[i][2] = x * sx;
					vtx[i][3] = y * sy;
					vtx[i][4] = b * sx;
					vtx[i][5] = a * sy;

					i++;
				}
			}
		}
	}

	nvtx = i;

	for (x = 0; x < max_planes; x++)
	{
		for (y = 0; y < max_polies; y++)
		{
			planeinfo[x].radiosity_setup[y] = 0;
		}
	}

	generating_radiosity = FALSE;
	

	texturesb = create_system_bitmap(16,16);
	texturesc = create_system_bitmap(16,16);
	for (y = 0; y < 16; y++)
	{
		for (x = 0; x < 16; x++)
		{				
			c = y * 16 + x;
			putpixel(texturesc, x, y, c);	
		}
	}

	sprintf(curfile, "");
}

void rotate_plane(int planenum)
{
	int	a;
	float	nx, ny, nz, x, y, z;

	for (a = 0; a <= planeinfo[planenum].lastpoint; a++)
	{
		x = planeinfo[planenum].pointinfo[a][0] - planeinfo[planenum].rbx;
		y = planeinfo[planenum].pointinfo[a][1] - planeinfo[planenum].rby;
		z = planeinfo[planenum].pointinfo[a][2] - planeinfo[planenum].rbz;

		rotpnt(x, y, z, planeinfo[planenum].xrot, planeinfo[planenum].yrot, planeinfo[planenum].zrot, &nx, &ny, &nz);

		planeinfo[planenum].pointinfo[a][3] = nx + planeinfo[planenum].rbx;
		planeinfo[planenum].pointinfo[a][4] = ny + planeinfo[planenum].rby;
		planeinfo[planenum].pointinfo[a][5] = nz + planeinfo[planenum].rbz;
	}

	calcsteps();
	setup_scene();
}

void rotate_point(int planenum)
{
	int	a, c, p;
	float	nx, ny, nz, x, y, z;

	for (c = 0; c < nselpoint; c++)
	{
		p = -1;
		if (selpoint[c][3] == -1)
		{
			for (a = 0; a <= planeinfo[planenum].lastpoint; a++)
			{
				x = planeinfo[planenum].pointinfo[a][3] + planeinfo[planenum].basex;
				y = planeinfo[planenum].pointinfo[a][4] + planeinfo[planenum].basey;
				z = planeinfo[planenum].pointinfo[a][5] + planeinfo[planenum].basez;
				if (checktolerance(selpoint[c][0], selpoint[c][1], selpoint[c][2], x, y, z))
				{
					p = a;
					continue;
				}
			}
		} else {
			p = selpoint[c][3];
		}
		
		if (p > -1)
		{
			x = planeinfo[planenum].pointinfo[p][0] - planeinfo[planenum].rbx;
			y = planeinfo[planenum].pointinfo[p][1] - planeinfo[planenum].rby;
			z = planeinfo[planenum].pointinfo[p][2] - planeinfo[planenum].rbz;

			rotpnt(x, y, z, planeinfo[planenum].xrot, planeinfo[planenum].yrot, planeinfo[planenum].zrot, &nx, &ny, &nz);

			planeinfo[planenum].pointinfo[p][3] = nx + planeinfo[planenum].rbx;
			planeinfo[planenum].pointinfo[p][4] = ny + planeinfo[planenum].rby;
			planeinfo[planenum].pointinfo[p][5] = nz + planeinfo[planenum].rbz;

			selpoint[c][0] = planeinfo[planenum].pointinfo[p][3] + planeinfo[planenum].basex;
			selpoint[c][1] = planeinfo[planenum].pointinfo[p][4] + planeinfo[planenum].basey;
			selpoint[c][2] = planeinfo[planenum].pointinfo[p][5] + planeinfo[planenum].basez;
		}
	}
	
	setup_scene();
}

void setup_plane(int planenum, short filltx)
{
	int	x, y, cols, rows, point, poly;
	float	yscale, scale[3], nx, ny, nz;

	scale[0] = (planeinfo[planenum].endx - planeinfo[planenum].basex) / planeinfo[planenum].cols;
	scale[1] = (planeinfo[planenum].endy - planeinfo[planenum].basey) / planeinfo[planenum].cols;
	scale[2] = (planeinfo[planenum].endz - planeinfo[planenum].basez) / planeinfo[planenum].cols;
	
	yscale = planeinfo[planenum].height / planeinfo[planenum].rows;

	cols = planeinfo[planenum].cols + 1;
	rows = planeinfo[planenum].rows + 1;

	planeinfo[planenum].lastpoint = 0;
	
	for (y = 0; y < rows; y++)
	{
		for (x = 0; x < cols; x++)
		{
			point = y * cols + x;
			planeinfo[planenum].pointinfo[point][0] = x * scale[0];
			planeinfo[planenum].pointinfo[point][1] = x * scale[1] + y * yscale;
			planeinfo[planenum].pointinfo[point][2] = x * scale[2];

			rotpnt(x * scale[0], x * scale[1] + y * yscale, x * scale[2], planeinfo[planenum].xrot, planeinfo[planenum].yrot, planeinfo[planenum].zrot, &nx, &ny, &nz);
			planeinfo[planenum].pointinfo[point][3] = nx;
			planeinfo[planenum].pointinfo[point][4] = ny;
			planeinfo[planenum].pointinfo[point][5] = nz;
			planeinfo[planenum].pointinfo[point][7] = 0;
			planeinfo[planenum].pointinfo[point][8] = 0;
			planeinfo[planenum].pointinfo[point][9] = 0;
			planeinfo[planenum].pointinfo[point][10] = 0;

			if (point > planeinfo[planenum].lastpoint) planeinfo[planenum].lastpoint = point;			
		}
	}

	poly = 0;

	planeinfo[planenum].lastpoly = 0;
	for (y = 0; y < rows-1; y++)
	{
		for (x = 0; x < cols-1; x++)
		{
			//y * cols + x
			planeinfo[planenum].polyinfo[poly][0] = (y+1) * cols + (x+0);
			planeinfo[planenum].polyinfo[poly][1] = (y+0) * cols + (x+0);
			planeinfo[planenum].polyinfo[poly][2] = (y+1) * cols + (x+1);			
			planeinfo[planenum].polyinfo[poly][3] = filltx;			
			planeinfo[planenum].polyinfo[poly][6] = 0;			
			planeinfo[planenum].polyvtx[poly][0] = vtx[0][0];
			planeinfo[planenum].polyvtx[poly][1] = vtx[0][1];
			planeinfo[planenum].polyvtx[poly][2] = vtx[0][2];
			planeinfo[planenum].polyvtx[poly][3] = vtx[0][3];
			planeinfo[planenum].polyvtx[poly][4] = vtx[0][4];
			planeinfo[planenum].polyvtx[poly][5] = vtx[0][5];
			planeinfo[planenum].polyinfo[poly][7] = 0;
			planeinfo[planenum].polyinfo[poly][8] = 0;
			planeinfo[planenum].polyinfo[poly][9] = 0;
			if (planeinfo[planenum].radiosity_setup[poly] == 2) planeinfo[planenum].radiosity_setup[poly] = 1;

			poly++;

			planeinfo[planenum].polyinfo[poly][0] = (y+0) * cols + (x+1);
			planeinfo[planenum].polyinfo[poly][1] = (y+1) * cols + (x+1);
			planeinfo[planenum].polyinfo[poly][2] = (y+0) * cols + (x+0);			
			planeinfo[planenum].polyinfo[poly][3] = filltx;
			planeinfo[planenum].polyinfo[poly][6] = 0;
			planeinfo[planenum].polyvtx[poly][0] = vtx[1][0];
			planeinfo[planenum].polyvtx[poly][1] = vtx[1][1];
			planeinfo[planenum].polyvtx[poly][2] = vtx[1][2];
			planeinfo[planenum].polyvtx[poly][3] = vtx[1][3];
			planeinfo[planenum].polyvtx[poly][4] = vtx[1][4];
			planeinfo[planenum].polyvtx[poly][5] = vtx[1][5];
			planeinfo[planenum].polyinfo[poly][7] = 0;
			planeinfo[planenum].polyinfo[poly][8] = 0;
			planeinfo[planenum].polyinfo[poly][9] = 0;
			if (planeinfo[planenum].radiosity_setup[poly] == 2) planeinfo[planenum].radiosity_setup[poly] = 1;

			if (poly > planeinfo[planenum].lastpoly) planeinfo[planenum].lastpoly = poly;
			poly++;
		}
	}

	int c;
	nplanes = 0;
	endplane = 0;

	for (c = 0; c < max_planes; c++)
		if (planeinfo[c].active) 
		{
			nplanes++;
			if (c > endplane) endplane = c;
		}

	calcsteps();
	setup_scene();
}


void setup_scene()
{	
	pointinfo.clear();
	polyinfo.clear();

	float nsx, nsy, nsz;

	int a, c, x , y, lightnum;

	point_t	tpoint;
	poly_t  tpoly;

	nlights = 0;

	for (c = 0; c <= endplane; c++)
	{		
		if (planeinfo[c].active)
		{			
			planeinfo[c].npoints = 0;
			planeinfo[c].npolies = 0;

			for (a = 0; a <= planeinfo[c].lastpoint; a++)
			{
				if (planeinfo[c].pointinfo[a][7] == 0)
				{
					nsx = planeinfo[c].pointinfo[a][3] + planeinfo[c].basex;
					nsy = planeinfo[c].pointinfo[a][4] + planeinfo[c].basey;
					nsz = planeinfo[c].pointinfo[a][5] + planeinfo[c].basez;

					tpoint.nsx = nsx * prexx + nsy * preyx + nsz * prezx;
					tpoint.nsy = nsx * prexy + nsy * preyy + nsz * prezy;
					tpoint.nsz = nsx * prexz + nsy * preyz + nsz * prezz;
					tpoint.plane = c;
					tpoint.point = a;
					tpoint.active = TRUE;

					planeinfo[c].pointinfo[a][6] = pointinfo.size();
					pointinfo.push_back(tpoint);
									
					planeinfo[c].npoints++;

					if (planeinfo[c].pointinfo[a][9] > 0 && planeinfo[c].pointinfo[a][10] == 0)
					{
						lightnum = planeinfo[c].pointinfo[a][9] - 1;
						lights[nlights][0] = planeinfo[c].pointinfo[a][6];
						lights[nlights][1] = lightnum;
						nlights++;
						if (lightnum > lastlight -1) lastlight = lightnum + 1;
					}

				}
			}

			for (a = 0; a <= planeinfo[c].lastpoly; a++)
			{
				if (planeinfo[c].polyinfo[a][6] < 2)
				{
					tpoly.p1 = planeinfo[c].pointinfo[planeinfo[c].polyinfo[a][0]][6];
					tpoly.p2 = planeinfo[c].pointinfo[planeinfo[c].polyinfo[a][1]][6];
					tpoly.p3 = planeinfo[c].pointinfo[planeinfo[c].polyinfo[a][2]][6];
					tpoly.tx = planeinfo[c].polyinfo[a][3];
					tpoly.u[0] = planeinfo[c].polyvtx[a][0];
					tpoly.v[0] = planeinfo[c].polyvtx[a][1];
					tpoly.u[1] = planeinfo[c].polyvtx[a][2];
					tpoly.v[1] = planeinfo[c].polyvtx[a][3];
					tpoly.u[2] = planeinfo[c].polyvtx[a][4];
					tpoly.v[2] = planeinfo[c].polyvtx[a][5];
										
					tpoly.plane = c;
					tpoly.poly = a;

									
					planeinfo[c].polyinfo[a][5] = polyinfo.size();
					if (planeinfo[c].radiosity_setup[a] == 0)
					{
						planeinfo[c].radiosity_setup[a] = 1;
						planeinfo[c].radiosity[a] = create_system_bitmap(16,16);						
					}
					if (planeinfo[c].radiosity_setup[a] == 1)
					{
						planeinfo[c].radiosity_setup[a] = 2;
						for (x = 0; x < 16; x++)
						{
							for (y = 0; y < 16; y++)
							{
								putpixel(planeinfo[c].radiosity[a], x, y, 0);
							}
						}						
						blit(textures[tpoly.tx], planeinfo[c].radiosity[a], 0, 0, 0, 0, 16, 16);
					}

					planeinfo[c].polyinfo[a][5] = polyinfo.size();
					polyinfo.push_back(tpoly);
					planeinfo[c].npolies++;
				}
			}
			
		}
	}

}

void surroundbox(short mode, short farleft, short fartop, short farright, short farbottom)
{
	short clearance, qx, qy;
	int whitec = makecol(192,192,192), span;

	span = 5;
	if (mode > 1) span = 1;

	clearance = span * xres / 320;
	if (mode == 1) whitec = makecol(0,0,192);
	if (mode == 3) whitec = makecol(0,192,0);

	qx = short((farright - farleft) * .25);
	qy = short((farbottom - fartop) * .25);

	line (offscreen, farleft - clearance, fartop - clearance, farleft - clearance, fartop - clearance + qy, whitec);
	line (offscreen, farleft - clearance, fartop - clearance, farleft - clearance + qx, fartop - clearance, whitec);

	line (offscreen, farright + clearance, fartop - clearance, farright + clearance, fartop - clearance + qy, whitec);
	line (offscreen, farright + clearance, fartop - clearance, farright + clearance - qx, fartop - clearance, whitec);
	
	line (offscreen, farright + clearance, farbottom + clearance, farright + clearance, farbottom + clearance - qy, whitec);
	line (offscreen, farright + clearance, farbottom + clearance, farright + clearance - qx, farbottom + clearance, whitec);

	line (offscreen, farleft - clearance, farbottom + clearance, farleft - clearance, farbottom + clearance - qy, whitec);
	line (offscreen, farleft - clearance, farbottom + clearance, farleft - clearance + qx, farbottom + clearance, whitec);
}

void updphysics()
{		
	focusofs[0] = focusofs[0] + (fightfocus[0] - focusofs[0]) * .08 * fpsr;
	focusofs[1] = focusofs[1] + (fightfocus[1] - focusofs[1]) * .08 * fpsr;
	focusofs[2] = focusofs[2] + (fightfocus[2] - focusofs[2]) * .08 * fpsr;
}


