/*
  Generation Olympia Engine - C++ Port
  Project Home: http://www.sourceforge.net/projects/qb2c/
  
  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.

  $Id: mapeditfileinput.cpp,v 1.6 2004/01/20 03:05:27 syn9 Exp $
*/

#include "../engine/fileinput.h"
#include "../engine/engineglobals.h"
#include "../engine/render.h"

#include "mapeditfileinput.h"
#include "mapeditglobals.h"
#include "mapeditgui.h"
#include "mapedit.h"

#include "../common/strutil.h"

void load_file(char* rfile)
{
	FILE *fp;
	char mapfile[100];
	int	c = 0;

	sprintf(mapfile, "data/maps/%s.dtl", rfile);
		
	if ((fp = fopen(mapfile, "r")) == NULL)
    {
        allegro_message("Error opening map file: %s\n", mapfile);
        return;
    }
		
	fscanf(fp, "%d", &c);

	fclose(fp);

	if (c == 1) load_file_v1(rfile);
	
}

void save_file_detailed(char* rfile)
{

	FILE *fp;
	char mapfile[100], tmp[100];
	int	c, d, e, whitec = makecol(192, 192, 192);

	sprintf(mapfile, "data/maps/%s.dtl", rfile);
		
	fp = fopen(mapfile, "w");
		
	int version = 1;

	fprintf(fp, "%d\n", version);

	fprintf(fp, "endplane\n");
	fprintf(fp, "%d\n", endplane);
	fprintf(fp, "%d\n", lastlight);
		
	fprintf(fp, "plane_info\n");

	int lx, ly, rx, ry, w, h;

	h = 16+5;
	w = 256;
	lx = xcenter - w / 2;
	ly = ycenter - h / 2;
	rx = lx + w - 1;
	ry = ly + h - 1;

	win (0, lx, ly, rx + 6, ry);
	
	sprintf(tmp, "Saving...");
	textout(offscreen, font, tmp, lx + 2, ly + 2, whitec);

	BITMAP *temppcx;

	updscreen();
	int greenc = makecol(0, 255, 0);

	for (c = 0; c <= endplane; c++)
	{
		rectfill (screen, lx + 2, ly + 10, lx + 2 + (w - 5) * (float)c / endplane, ly + 18, greenc);
		if (planeinfo[c].active) fprintf(fp, "1\n");
		if (!planeinfo[c].active) fprintf(fp, "0\n");
		if (planeinfo[c].condensed) fprintf(fp, "1\n");
		if (!planeinfo[c].condensed) fprintf(fp, "0\n");

		fprintf(fp, "%f\n", planeinfo[c].basex);
		fprintf(fp, "%f\n", planeinfo[c].basey);
		fprintf(fp, "%f\n", planeinfo[c].basez);

		fprintf(fp, "%f\n", planeinfo[c].xrot);
		fprintf(fp, "%f\n", planeinfo[c].yrot);
		fprintf(fp, "%f\n", planeinfo[c].zrot);

		fprintf(fp, "%f\n", planeinfo[c].endx);
		fprintf(fp, "%f\n", planeinfo[c].endy);
		fprintf(fp, "%f\n", planeinfo[c].endz);

		fprintf(fp, "%f\n", planeinfo[c].height);

		fprintf(fp, "%f\n", planeinfo[c].rbx);
		fprintf(fp, "%f\n", planeinfo[c].rby);
		fprintf(fp, "%f\n", planeinfo[c].rbz);

		fprintf(fp, "%d\n", planeinfo[c].rows);
		fprintf(fp, "%d\n", planeinfo[c].cols);
		fprintf(fp, "%d\n", planeinfo[c].lastpoint);
		fprintf(fp, "%d\n", planeinfo[c].lastpoly);
		fprintf(fp, "%d\n", planeinfo[c].npoints);
		fprintf(fp, "%d\n", planeinfo[c].npolies);

		if (planeinfo[c].active) {

			for (d = 0; d <= planeinfo[c].lastpoint; d++)
			{
				for (e = 0; e < 11; e++)
				{
					fprintf(fp, "%f\n", planeinfo[c].pointinfo[d][e]);
				}
			}

			temppcx = create_system_bitmap((planeinfo[c].lastpoly + 1) * 16, 16);
			clear(temppcx);

			for (d = 0; d <= planeinfo[c].lastpoly; d++)
			{
				for (e = 0; e < 10; e++)
				{
					fprintf(fp, "%d\n", planeinfo[c].polyinfo[d][e]);
				}
				for (e = 0; e < 6; e++)
				{
					fprintf(fp, "%d\n", planeinfo[c].polyvtx[d][e]);
				}

				e = 0;
				if (planeinfo[c].radiosity_setup[d] > 0) e = 1;
				fprintf(fp, "%d\n", e);

			
				if (planeinfo[c].radiosity_setup[d] > 0)
				{
					blit(planeinfo[c].radiosity[d], temppcx, 0, 0, d * 16, 0, 16, 16);
					//sprintf(tmp, "data/maps/shadowmap/i%s_%d_%d.bmp", rfile, c, d);
					//save_bmp(tmp, planeinfo[c].radiosity[d], pal);
				}
			}
			sprintf(tmp, "data/maps/shadowmap/i%s_%d.bmp", rfile, c);
			save_bmp(tmp, temppcx, pal);

			destroy_bitmap(temppcx);

		}		
		
		for (d = 0; d < lastlight; d++)
		{
			fprintf(fp, "%f\n", lightinfo[d].strength);
			fprintf(fp, "%d\n", lightinfo[d].lightr);
			fprintf(fp, "%d\n", lightinfo[d].lightg);
			fprintf(fp, "%d\n", lightinfo[d].lightb);
		}
		
	}

	fclose(fp);
}

void save_file_simple(char* rfile)
{

	FILE *fp;
	char mapfile[100];

	sprintf(mapfile, "data/maps/%s.smp", rfile);
		
	fp = fopen(mapfile, "w");
		
	int version = 1;

	fprintf(fp, "%d\n", version);

	fclose(fp);
}


