/*
  Generation Olympia Engine - C++ Port
  Project Home: http://www.sourceforge.net/projects/qb2c/
  
  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.

  $Id: render.cpp,v 1.13 2004/01/20 02:48:43 syn9 Exp $
*/

#include "engineglobals.h"
#include "render.h"
#include "fileinput.h"

#include <math.h>

void calcpoints()
{
	short sleft, sright, sup, sdown;
    float nsx, nsy, nsz, sx, sy, sz;
	float psx, psy, psz, wsx, wsy, wsz, d, xx, yy, zz;
	unsigned int i;

	short xc, yc, scy, xr, yr;
	float xzr, yzr;

	xc = xcenter;
	yc = ycenter;
	xr = xres;
	yr = yres;
	xzr = xzoomratio;
	yzr = yzoomratio;
	scy = scrny;

	sleft = -xc  / 2;
	sright = xr + xc  / 2;
	sup = 10 * yres / 240;
	sdown = yr;

	npoints = 0;

	if (generating_radiosity)
	{
		xc = lightres / 2;
		yc = lightres / 2;
		xr = lightres;
		yr = lightres;

		sleft = -xc * 2;
		sright = xr + xc * 2;
		sup = -yc * 2;
		sdown = yr + yc * 2;
	}
	
	nsx = -focusofs[0];
	nsy = -focusofs[1];
	nsz = -focusofs[2];

	psx = nsx * prexx + nsy * preyx + nsz * prezx;
	psy = nsx * prexy + nsy * preyy + nsz * prezy;
	psz = nsx * prexz + nsy * preyz + nsz * prezz;
		
	for (i = 0; i != pointinfo.size(); i++)
	{
		if (!pointinfo[i].active) continue;

		wsx = 0;
		wsy = 0;
		wsz = 0;
		if (pointinfo[i].water == 1)
		{
			xx = pointinfo[i].xx - ripplex;
			zz = pointinfo[i].zz - ripplez;

			d = sqrt(xx * xx + zz * zz);
			nsx = 0;
			nsy = sin(rad * fpr / FPS * 720 + d / 4 * 360) * rippleamp * (1 - d / ripplefalloff);
			nsz = 0;

			wsx = nsx * prexx + nsy * preyx + nsz * prezx;
			wsy = nsx * prexy + nsy * preyy + nsz * prezy;
			wsz = nsx * prexz + nsy * preyz + nsz * prezz;

			pointinfo[i].yy = pointinfo[i].ybase + nsy;
		}
		if (pointinfo[i].water == 2)
		{
			yy = pointinfo[i].yy;

			nsx = cos(rad * fpr / FPS * 720 + yy / 4 * 360) * rippleamp * .5;
			nsy = 0;
			nsz = sin(rad * fpr / FPS * 720 + yy / 4 * 360) * rippleamp * .5;

			wsx = nsx * prexx + nsy * preyx + nsz * prezx;
			wsy = nsx * prexy + nsy * preyy + nsz * prezy;
			wsz = nsx * prexz + nsy * preyz + nsz * prezz;

			pointinfo[i].yy = pointinfo[i].ybase + nsy;
		}


		nsx = pointinfo[i].nsx;
		nsy = pointinfo[i].nsy;
		nsz = pointinfo[i].nsz;

		sx = nsx + wsx + psx - camx;
		sy = nsy + wsy + psy - camy;
		sz = nsz + wsz + psz - camz;

		pointinfo[i].sz = sz;

		pointinfo[i].psx = sx + vdist;
		pointinfo[i].psy = sy + vdist;
		pointinfo[i].psz = sz + vdist;
			
		pointinfo[i].flag = 1;
		
		if (sz == 0) sz = -.1;
		
		if (sz > 0 && !generating_radiosity) continue;
		if (sz > 0 && generating_radiosity) continue;
		
		pointinfo[i].x = xc + sx / sz * xzr;
		pointinfo[i].y = yc + sy / sz * yzr - scy;
		
		if (pointinfo[i].x < sleft || pointinfo[i].x > sright) continue;
		if (pointinfo[i].y < sup || pointinfo[i].y > sdown) continue;		
		
		pointinfo[i].scale = -sz * 75;
		
		pointinfo[i].flag = 0;
		
		npoints++;
	}
}




void calcsteps()
{
    // Rel's performance addition to the QB engine (which seems to be unnecessary given
    // this post (http://forum.qbasicnews.com/viewtopic.php?p=46965#46965) by syn9)
    //Rotmatrix(camrx, camry, camrz);

    float stpc = stp;
	
	//if (generating_radiosity) stp = 2 * stp;
	rotpnt(-stp, 0, 0, camrx, camry, camrz, &prexx, &prexy, &prexz);
    rotpnt(0, stp * .8, 0, camrx, camry, camrz, &preyx, &preyy, &preyz);
    rotpnt(0, 0, stp, camrx, camry, camrz, &prezx, &prezy, &prezz);

	stp = stpc;
}




// Update the fading state of the screen if necessary
void fade()
{
	if (current_fade_level != target_fade_level)
	{
		if (current_fade_level > target_fade_level)
		{
			// Fade out
			current_fade_level -= fade_speed;
			if (current_fade_level < target_fade_level)
				current_fade_level = target_fade_level;
		}
		else if (current_fade_level < target_fade_level)
		{
			// Fade in
			current_fade_level += fade_speed;
			if (current_fade_level > target_fade_level)
				current_fade_level = target_fade_level;
		}
    }

	if (current_fade_level < 1.0 && current_fade_level >= 0.0)
	{
		set_trans_blender(0,0,0, 255 - int(current_fade_level * 255));
		drawing_mode(DRAW_MODE_TRANS, NULL, 0, 0);
		rectfill(offscreen, 0, 0, SCREEN_W, SCREEN_H, 0);
		drawing_mode(DRAW_MODE_SOLID, NULL, 0, 0);
	}
}


short getpoint(float nsx, float nsy, float nsz)
{
    float tolerance = 0;
    short p, found = -1;
    short temppoints = 0;
	unsigned int i;

    for (i = 0; i != pointinfo.size(); i++)
    {
        p = 0;
        if (nsx >= pointinfo[i].nsx - tolerance && nsx <= pointinfo[i].nsx + tolerance)
            p = p + 1;
        if (nsy >= pointinfo[i].nsy - tolerance && nsy <= pointinfo[i].nsy + tolerance)
            p = p + 1;
        if (nsz >= pointinfo[i].nsz - tolerance && nsz <= pointinfo[i].nsz + tolerance)
            p = p + 1;
        if (p == 3)
            found = i;
    }

    if (found > -1)
        return temppoints;

	point_t point = { nsx, nsy, nsz};
	pointinfo.push_back( point);
    return 0;
}

double getx(float a, float b)
{
    double d;

    if (b >= -.1)
        b = -.1;
    d = (xcenter + (a / b * xzoomratio));

    if (d < -500)
        d = -500;
    if (d > 820)
        d = 820;

    return(d);
}

double gety(float a, float b)
{
    double d;

    if (b >= -.1)
        b = -.1;

    d = (ycenter + (a / b * yzoomratio));

    if (d < -500)
        d = -500;
    if (d > 700)
        d = 700;

    return (d - scrny);
}

void initialize(short bits, short mode, short xsize, short ysize)
{
    short i;
    //BITMAP* temppcx;

    srand((unsigned)time(NULL));


	xres = xsize;
	yres = ysize;
	int gfx_mode;

	if (mode == 0) gfx_mode = GFX_AUTODETECT_WINDOWED;
	if (mode == 1) gfx_mode = GFX_AUTODETECT_FULLSCREEN;

	xcenter = xres / 2;
	ycenter = yres / 2;

	xzoomratio = xres / 1.6;
	yzoomratio = xres / 1.6;

	stp = 7;

	printf("Initalizing Allegro...\n");

	allegro_init();
	install_keyboard();

	printf("Setting graphics mode...\n");

	set_color_depth(bits);
	if(set_gfx_mode(gfx_mode, xres, yres, 0, 0) != 0)
	{ 
		set_color_depth(16);
		if(set_gfx_mode(gfx_mode, xres, yres, 0, 0) != 0)
		{ 
			set_color_depth(24);
			if(set_gfx_mode(gfx_mode, xres, yres, 0, 0) != 0)
			{ 
				set_color_depth(32);
				if(set_gfx_mode(gfx_mode, xres, yres, 0, 0) != 0)
				{ 
					set_color_depth(15);
					if(set_gfx_mode(gfx_mode, xres, yres, 0, 0) != 0)
					{ 
						allegro_message("Can't init graphics window!"); 
						exit(1);
					}
				} 
			}
		}
	}

	//drawing_mode(DRAW_MODE_TRANS, NULL, 0, 0);	

	bfc = TRUE;

    refreshrate = get_refresh_rate();
    if (refreshrate == 0)
        refreshrate = 60;

	offscreen = create_system_bitmap(xres, yres);
	temp2 = create_system_bitmap(31, 31);
	temp = create_system_bitmap(23, 36);
	if (!temp2 || !temp || !offscreen) {
		printf("Error while allocating bitmaps!\n");
		exit(1);
	}
	for (i = 0; i < 300; i++) {
		textures[i] = create_system_bitmap(16, 16);
		if (!textures[i]) {
			printf("Error while allocating bitmaps!\n");
			exit(1);
		}
	}
	
	
	textout(screen, font, "Initializing", 0, 0, 7);
	
	camx = 0;
	camy = -4.5;
	camz = 55;
	
	//camrx = 50;
	camrx = 45;
	camry = 0;
	camrz = 0;

	gamma = 1;

	scrny = short(-12 * xres / 320);

	text_mode(-1);

}

void deinit()
{
    remove_keyboard();
    allegro_exit();
}


// Rel's performance additions to the QB engine
void Rotmatrix(float AngleX, float AngleY, float AngleZ)
{
    float rAnglex, rAngley, rAnglez, cx, cy, cz, sx, sy, sz;

    rAnglex = AngleX * rad;
    rAngley = AngleY * rad;
    rAnglez = AngleZ * rad;

    //Precalculate the SIN and COS of each angle
    cx = cos(rAnglex);
    sx = sin(rAnglex);
    cy = cos(rAngley);
    sy = sin(rAngley);
    cz = cos(rAnglez);
    sz = sin(rAnglez);

    //Transformation matrix formula

    tx1 = cy * cz;
    tx2 = cy * sz;
    tx3 = -sy;
    ty1 = cx * -sz + sx * sy * cz;
    ty2 = cx * cz + sx * sy * sz;
    ty3 = sx * cy;
    tz1 = -sx * -sz + cx * sy * cz;
    tz2 = -sx * cz + cz * sy * sz;
    tz3 = cx * cy;
}

void rotpnt(float px, float py, float pz, float xr, float yr, float zr, float* newx, float* newy, float* newz)
{

	rx0 = px;
    ry0 = py;
    rz0 = pz;
                       
    rx1 = rx0 * cos(rad * zr) - ry0 * sin(rad * zr);
    ry1 = rx0 * sin(rad * zr) + ry0 * cos(rad * zr);
    rz1 = rz0;

    ry2 = ry1 * cos(rad * xr) - rz1 * sin(rad * xr);
    rz2 = ry1 * sin(rad * xr) + rz1 * cos(rad * xr);
    rx2 = rx1;

    rz3 = rz2 * cos(rad * yr) - rx2 * sin(rad * yr);
    rx3 = rz2 * sin(rad * yr) + rx2 * cos(rad * yr);
    ry3 = ry2;

    *newx = rx3;
    *newy = ry3;
    *newz = rz3;
}

void sortpolys()
{
	int i = 0, p1, p2, p3;
	float nsx, nsy, nsz, psx, psy, psz;
	short x1, y1, x2, y2, x3, y3;
	float znormal;

	nsx = -focusofs[0];
	nsy = -focusofs[1];
	nsz = -focusofs[2];

	psx = nsx * prexx + nsy * preyx + nsz * prezx;
	psy = nsx * prexy + nsy * preyy + nsz * prezy;
	psz = nsx * prexz + nsy * preyz + nsz * prezz;
	

    for(std::vector<poly_t>::const_iterator itr = polyinfo.begin(); itr != polyinfo.end(); ++itr, ++i)
    {
        bool b  = pointinfo[ itr->p1].flag | pointinfo[ itr->p2].flag | pointinfo[ itr->p3].flag;

        if(!b)
		{
			p1 = polyinfo[i].p1;
			p2 = polyinfo[i].p2;
			p3 = polyinfo[i].p3;
			
			
			if (bfc)
			{
				x1 = short(pointinfo[p1].x);
				y1 = short(pointinfo[p1].y);
				x2 = short(pointinfo[p2].x);
				y2 = short(pointinfo[p2].y);
				x3 = short(pointinfo[p3].x);
				y3 = short(pointinfo[p3].y);
			
				znormal = (x2 - x1) * (y1 - y3) - (y2 - y1) * (x1 - x3);
				if (znormal < 0) continue;
			}
			
			if (generating_radiosity && (pointinfo[p1].sz > 0 && pointinfo[p2].sz > 0 && pointinfo[p3].sz > 0)) continue;
			if (generating_radiosity && (pointinfo[p1].x < 0 && pointinfo[p2].x < 0 && pointinfo[p3].x < 0)) continue;
			if (generating_radiosity && (pointinfo[p1].x > xres && pointinfo[p2].x > xres && pointinfo[p3].x > xres)) continue;
			if (generating_radiosity && (pointinfo[p1].y < 0 && pointinfo[p2].y < 0 && pointinfo[p3].y < 0)) continue;
			if (generating_radiosity && (pointinfo[p1].y > yres && pointinfo[p2].y > yres && pointinfo[p3].y > yres)) continue;


			int ind = (int)(pointinfo[ itr->p1].scale + pointinfo[ itr->p2].scale + pointinfo[ itr->p3].scale) / 3;
			
			for(;;)
            {
                if (ind > vdist || ind < 0)
                    break;

                if (dbuffer[ind] > 0)
                    ind = ind + 1;
                if (dbuffer[ind] == 0)
                {
                    dbuffer[ind] = i + 100;
    
                    if (ind < start)
                        start = ind;
                    if (ind > depth)
                        depth = ind;
                    break;
                }
            }
        }
    }
}

float u(float deg)
{
    float d;

    d = deg;
    if (d > 359)
        d = d - 360;
    if (d < 0)
        d = d + 360;

    return(d);
}

void updscreen(bool s)
{		
    
	fpr++;
	if (globaltime < clock())
	{
		FPS = fpr;
		fpr = 0;
		fpsr = 6 / FPS;
		globaltime = clock() + 100;
	}
	timer = timer + .001 * fpsr;
	
	if (s)
	{
		show_mouse(offscreen);
		blit(offscreen, screen, 0, 0, 0, 0, SCREEN_W, SCREEN_H);		
		show_mouse(NULL);
	}
}

short valid(short x, short y)
{
    short top, bot;

    top = 30 - 42 - scrny;
    bot = 130 - 42 - scrny;

    if (x > 15 && x < 304 && y > top && y < bot)
        return(1);
    else
        return(0);
}	

void win(short l, short x1, short y1, short x2, short y2)
{
	int mgreyc, lgreyc;

	
	lgreyc = makecol(50, 97, 153);
	mgreyc = makecol(33, 64, 102);
	if (l == 1) mgreyc = makecol(192,192,192);	

	rectfill (offscreen, x1, y1, x2 - 6, y2, lgreyc);

	if (l < 4)
	{		
		hline (offscreen, x1, y1, x2 - 6, mgreyc);
		hline (offscreen, x1, y2, x2, mgreyc);
		vline (offscreen, x1, y1, y2, mgreyc);
	
		int c;
		for (c = 0; c < 7; c++)
		{
			vline (offscreen, x2 - 6 + c, y1 + c, y2, mgreyc);
		}
	} else {
		hline (offscreen, x1, y2, x2 - 6, mgreyc);
		hline (offscreen, x1, y1, x2, mgreyc);
		vline (offscreen, x1, y1, y2, mgreyc);
	
		int c;
		for (c = 0; c < 7; c++)
		{
			vline (offscreen, x2 - 6 + c, y1, y2 - c, mgreyc);
		}
	}
	
}


